from enum import Enum
from typing import Any, Protocol

from jax import Array

# Many JAX functions are designed to work with scalar numerical values. This also
# includes zero dimensional jax arrays.
Scalar = int | float | Array


ParamsDict = dict[str, Any]


class UserFunction(Protocol):
    """A function provided by the user.

    Only used for type checking.

    """

    def __call__(self, *args: Any, **kwargs: Any) -> Any: ...  # noqa: ANN401, D102


class InternalUserFunction(Protocol):
    """The internal representation of a function provided by the user.

    Only used for type checking.

    """

    def __call__(  # noqa: D102
        self, *args: Scalar, params: ParamsDict, **kwargs: Scalar
    ) -> Scalar: ...


class MaxQOverCFunction(Protocol):
    """The function that maximizes Q over the continuous actions.

    Q is the state-action value function. The MaxQOverCFunction returns the maximum of Q
    over the continuous actions.

    Only used for type checking.

    """

    def __call__(  # noqa: D102
        self, next_V_arr: Array, params: ParamsDict, **kwargs: Scalar
    ) -> Array: ...


class ArgmaxQOverCFunction(Protocol):
    """The function that finds the argmax of Q over the continuous actions.

    Q is the state-action value function. The ArgmaxQOverCFunction returns the argmax
    and the maximum of Q over the continuous actions.

    Only used for type checking.

    """

    def __call__(  # noqa: D102
        self, next_V_arr: Array, params: ParamsDict, **kwargs: Scalar
    ) -> tuple[Array, Array]: ...


class MaxQcOverDFunction(Protocol):
    """The function that maximizes Qc over the discrete actions.

    Qc is the maximum of the state-action value function (Q) over the continuous
    actions, conditional on the discrete action. It depends on a state and the discrete
    actions. The MaxQcFunction returns the maximum of Qc over the discrete actions.

    Only used for type checking.

    """

    def __call__(self, Qc_arr: Array, params: ParamsDict) -> Array: ...  # noqa: D102


class ArgmaxQcOverDFunction(Protocol):
    """The function that finds the argmax of Qc over the discrete actions.

    Qc is the maximum of the state-action value function (Q) over the continuous
    actions, conditional on the discrete action. It depends on a state and the discrete
    actions. The ArgmaxQcFunction returns the argmax of Qc over the discrete actions.

    Only used for type checking.

    """

    def __call__(self, Qc_arr: Array, params: ParamsDict) -> tuple[Array, Array]: ...  # noqa: D102


class StochasticNextFunction(Protocol):
    """The function that simulates the next state of a stochastic variable.

    Only used for type checking.

    """

    def __call__(self, keys: dict[str, Array], **kwargs: Array) -> Array: ...  # noqa: D102


class ShockType(Enum):
    """Type of shocks."""

    EXTREME_VALUE = "extreme_value"
    NONE = None


class Target(Enum):
    """Target of the function."""

    SOLVE = "solve"
    SIMULATE = "simulate"



"""Create a state space for a given model."""

import pandas as pd
from jax import Array

from lcm.grids import ContinuousGrid, DiscreteGrid
from lcm.interfaces import InternalModel, StateActionSpace, StateSpaceInfo


def create_state_action_space(
    model: InternalModel,
    *,
    initial_states: dict[str, Array] | None = None,
    is_last_period: bool = False,
) -> StateActionSpace:
    """Create a state-action-space.

    Creates the state-action-space for the solution and simulation of a model. In the
    simulation, initial states must be provided.

    Args:
        model: A processed model.
        initial_states: A dictionary with the initial values of the state variables.
            If None, the initial values are the minimum values of the state variables.
        is_last_period: Whether the state-action-space is created for the last period,
            in which case auxiliary variables are not included.

    Returns:
        A state-action-space. Contains the grids of the discrete and continuous actions,
        the grids of the state variables, or the initial values of the state variables,
        and the names of the state and action variables in the order they appear in the
        variable info table.

    """
    vi = model.variable_info
    if is_last_period:
        vi = vi.query("~is_auxiliary")

    if initial_states is None:
        states = {sn: model.grids[sn] for sn in vi.query("is_state").index}
    else:
        _validate_initial_states_names(initial_states, variable_info=vi)
        states = initial_states

    discrete_actions = {
        name: model.grids[name] for name in vi.query("is_action & is_discrete").index
    }
    continuous_actions = {
        name: model.grids[name] for name in vi.query("is_action & is_continuous").index
    }
    ordered_var_names = tuple(vi.query("is_state | is_discrete").index)

    return StateActionSpace(
        states=states,
        discrete_actions=discrete_actions,
        continuous_actions=continuous_actions,
        states_and_discrete_actions_names=ordered_var_names,
    )


def create_state_space_info(
    model: InternalModel,
    *,
    is_last_period: bool,
) -> StateSpaceInfo:
    """Collect information on the state space for the model solution.

    A state-space information is a compressed representation of all feasible states.

    Args:
        model: A processed model.
        is_last_period: Whether the function is created for the last period.

    Returns:
        The state-space information.

    """
    vi = model.variable_info
    if is_last_period:
        vi = vi.query("~is_auxiliary")

    state_names = vi.query("is_state").index.tolist()

    discrete_states = {
        name: grid_spec
        for name, grid_spec in model.gridspecs.items()
        if name in state_names and isinstance(grid_spec, DiscreteGrid)
    }

    continuous_states = {
        name: grid_spec
        for name, grid_spec in model.gridspecs.items()
        if name in state_names and isinstance(grid_spec, ContinuousGrid)
    }

    return StateSpaceInfo(
        states_names=tuple(state_names),
        discrete_states=discrete_states,
        continuous_states=continuous_states,
    )


def _validate_initial_states_names(
    initial_states: dict[str, Array], variable_info: pd.DataFrame
) -> None:
    """Checks if each model-state has an initial value."""
    states_names_in_model = set(variable_info.query("is_state").index)
    provided_states_names = set(initial_states)

    if states_names_in_model != provided_states_names:
        missing = states_names_in_model - provided_states_names
        too_many = provided_states_names - states_names_in_model
        raise ValueError(
            "You need to provide an initial array for each state variable in the model."
            f"\n\nMissing initial states: {missing}\n",
            f"Provided variables that are not states: {too_many}",
        )



import functools
import inspect
from collections.abc import Callable
from typing import Any, TypeVar, cast

from lcm.typing import Scalar

ReturnType = TypeVar("ReturnType")


def allow_only_kwargs(func: Callable[..., ReturnType]) -> Callable[..., ReturnType]:
    """Restrict a function to be called with only keyword arguments.

    Args:
        func: The function to be wrapped.

    Returns:
        A Callable with the same arguments as func (but with the additional restriction
            that it is only callable with keyword arguments).

    """
    signature = inspect.signature(func)
    parameters = signature.parameters

    # Get names of keyword-only arguments
    kw_only_parameters = [
        p.name for p in parameters.values() if p.kind == inspect.Parameter.KEYWORD_ONLY
    ]

    # Create new signature without positional-only arguments
    new_parameters = [
        p.replace(kind=inspect.Parameter.KEYWORD_ONLY) for p in parameters.values()
    ]
    new_signature = signature.replace(parameters=new_parameters)

    @functools.wraps(func)
    def func_with_only_kwargs(*args: Scalar, **kwargs: Scalar) -> ReturnType:
        if args:
            raise ValueError(
                (
                    "This function has been decorated so that it allows only kwargs, "
                    "but was called with positional arguments."
                ),
            )

        extra = set(kwargs).difference(parameters)
        if extra:
            raise ValueError(
                f"Expected arguments: {list(parameters)}, got extra: {extra}",
            )

        missing = set(parameters).difference(kwargs)
        if missing:
            raise ValueError(
                f"Expected arguments: {list(parameters)}, missing: {missing}",
            )

        # Retrieve keyword-only arguments
        kw_only_kwargs = {k: kwargs[k] for k in kw_only_parameters}

        # Get kwargs that must be converted to positional arguments
        pos_kwargs = {k: v for k, v in kwargs.items() if k not in kw_only_parameters}

        # Collect all positional arguments in correct order
        positional = convert_kwargs_to_args(pos_kwargs, list(parameters))

        return func(*positional, **kw_only_kwargs)

    # This raises a mypy error but is perfectly fine to do. See
    # https://github.com/python/mypy/issues/12472
    func_with_only_kwargs.__signature__ = new_signature  # type: ignore[attr-defined]

    # We cast to F here to signal mypy that the return type is the same as the input
    # type. This ignores the change of parameters from positional to keyword-only
    # arguments.
    # TODO(@timmens): Remove this cast once we find an explicit way to specify the
    # change from positional to keyword-only parameter in the function signature
    # https://github.com/opensourceeconomics/pylcm/issues/80.
    return cast(Callable[..., ReturnType], func_with_only_kwargs)


def allow_args(func: Callable[..., ReturnType]) -> Callable[..., ReturnType]:
    """Allow a function to be called with positional arguments.

    In comparison to `allow_only_kwargs`, the `allow_args` decorator does not enforce
    that the function is called only with positional arguments.

    Args:
        func: The function to be wrapped.

    Returns:
        A Callable with the same arguments as func (but with the additional possibility
            to call it with positional arguments).

    """
    signature = inspect.signature(func)
    parameters = signature.parameters

    # Count the number of positional-only arguments
    n_positional_only_parameters = len(
        [p for p in parameters.values() if p.kind == inspect.Parameter.POSITIONAL_ONLY],
    )

    # Create new signature without keyword-only arguments
    new_parameters = [
        (
            p.replace(kind=inspect.Parameter.POSITIONAL_OR_KEYWORD)
            if p.kind == inspect.Parameter.KEYWORD_ONLY
            else p
        )
        for p in parameters.values()
    ]
    new_signature = signature.replace(parameters=new_parameters)

    @functools.wraps(func)
    def allow_args_wrapper(*args: Scalar, **kwargs: Scalar) -> ReturnType:
        # Check if the total number of arguments matches the function signature
        if len(args) + len(kwargs) != len(parameters):
            too_many = len(args) + len(kwargs) > len(parameters)
            msg = (
                "Too many arguments provided."
                if too_many
                else "Not all arguments provided."
            )
            raise ValueError(msg)

        # Convert all arguments to positional arguments in correct order
        positional = list(args) + convert_kwargs_to_args(kwargs, list(parameters))

        # Extract positional-only arguments
        positional_only = positional[:n_positional_only_parameters]

        # Create kwargs dictionary with remaining arguments
        kwargs_names = list(parameters)[n_positional_only_parameters:]
        kwargs = dict(
            zip(
                kwargs_names,
                positional[n_positional_only_parameters:],
                strict=True,
            ),
        )

        return func(*positional_only, **kwargs)

    # This raises a mypy error but is perfectly fine to do. See
    # https://github.com/python/mypy/issues/12472
    allow_args_wrapper.__signature__ = new_signature  # type: ignore[attr-defined]

    # We cast to F here to signal mypy that the return type is the same as the input
    # type. This ignores the change of parameters from positional to keyword-only
    # arguments.
    # TODO(@timmens): Remove this cast once we find an explicit way to specify the
    # change from positional to keyword-only parameter in the function signature
    # https://github.com/opensourceeconomics/pylcm/issues/80.
    return cast(Callable[..., ReturnType], allow_args_wrapper)


def get_union_of_arguments(list_of_functions: list[Callable[..., Any]]) -> set[str]:
    """Return the union of arguments of a list of functions.

    Args:
        list_of_functions: A list of functions.

    Returns:
        The union of arguments of all functions in list_of_functions.

    """
    arguments = [inspect.signature(f).parameters for f in list_of_functions]
    return set().union(*arguments)


def all_as_kwargs(
    args: tuple[Any, ...],
    kwargs: dict[str, Any],
    arg_names: list[str],
) -> dict[str, Any]:
    """Return kwargs dictionary containing all arguments.

    Args:
        args: Positional arguments.
        kwargs: Keyword arguments.
        arg_names: Names of arguments.

    Returns:
        A dictionary of all arguments.

    """
    return dict(zip(arg_names[: len(args)], args, strict=True)) | kwargs


def all_as_args(
    args: tuple[Any, ...],
    kwargs: dict[str, Any],
    arg_names: list[str],
) -> tuple[Any, ...]:
    """Return args tuple containing all arguments.

    Args:
        args: Positional arguments.
        kwargs: Keyword arguments.
        arg_names: Names of arguments.

    Returns:
        A tuple of all arguments.

    """
    return args + tuple(convert_kwargs_to_args(kwargs, arg_names))


def convert_kwargs_to_args(kwargs: dict[str, Any], parameters: list[str]) -> list[Any]:
    """Convert kwargs to args in the order of parameters.

    Args:
        kwargs: Keyword arguments.
        parameters: List of parameter names in the order they should be.

    Returns:
        List of arguments in the order of parameters.

    """
    sorted_kwargs = dict(sorted(kwargs.items(), key=lambda kw: parameters.index(kw[0])))
    return list(sorted_kwargs.values())



try:
    import pdbp  # noqa: F401
except ImportError:
    pass

from lcm import mark
from lcm.grids import DiscreteGrid, LinspaceGrid, LogspaceGrid
from lcm.user_model import Model

__all__ = ["DiscreteGrid", "LinspaceGrid", "LogspaceGrid", "Model", "mark"]



"""Collection of LCM marking decorators."""

import functools
from collections.abc import Callable
from dataclasses import dataclass
from typing import Any, ParamSpec, TypeVar

P = ParamSpec("P")
R = TypeVar("R")


@dataclass(frozen=True)
class StochasticInfo:
    """Information on the stochastic nature of user provided functions."""


def stochastic(
    func: Callable[..., Any],
    *args: tuple[Any, ...],
    **kwargs: dict[str, Any],
) -> Callable[..., Any]:
    """Decorator to mark a function as stochastic and add information.

    Args:
        func (callable): The function to be decorated.
        *args (list): Positional arguments to be passed to the StochasticInfo.
        **kwargs (dict): Keyword arguments to be passed to the StochasticInfo.

    Returns:
        The decorated function

    """
    stochastic_info = StochasticInfo(*args, **kwargs)

    def decorator_stochastic(func: Callable[P, R]) -> Callable[P, R]:
        @functools.wraps(func)
        def wrapper_mark_stochastic(*args: P.args, **kwargs: P.kwargs) -> R:
            return func(*args, **kwargs)

        wrapper_mark_stochastic._stochastic_info = stochastic_info  # type: ignore[attr-defined]
        return wrapper_mark_stochastic

    return decorator_stochastic(func) if callable(func) else decorator_stochastic



"""Create a parameter template for a model specification."""

import inspect

import jax.numpy as jnp
import pandas as pd
from jax import Array

from lcm.input_processing.util import get_grids, get_variable_info
from lcm.typing import ParamsDict
from lcm.user_model import Model


def create_params_template(
    model: Model,
    default_params: dict[str, float] = {"beta": jnp.nan},  # noqa: B006
) -> ParamsDict:
    """Create parameter template from a model specification.

    Args:
        model: The model as provided by the user.
        default_params: A dictionary of default parameters. Default is None. If None,
            the default {"beta": np.nan} is used. For other lifetime reward objectives,
            additional parameters may be required, for example {"beta": np.nan, "delta":
            np.nan} for beta-delta discounting.

    Returns:
        A nested dictionary of model parameters.

    """
    variable_info = get_variable_info(model)
    grids = get_grids(model)

    if variable_info["is_stochastic"].any():
        stochastic_transitions = _create_stochastic_transition_params(
            model=model,
            variable_info=variable_info,
            grids=grids,
        )
        stochastic_transition_params = {"shocks": stochastic_transitions}
    else:
        stochastic_transition_params = {}

    function_params = _create_function_params(model)

    return default_params | function_params | stochastic_transition_params


def _create_function_params(model: Model) -> dict[str, dict[str, float]]:
    """Get function parameters from a model specification.

    Explanation: We consider the arguments of all model functions, from which we exclude
    all variables that are states, actions or the period argument. Everything else is
    considered a parameter of the respective model function that is provided by the
    user.

    Args:
        model: The model as provided by the user.

    Returns:
        A dictionary for each model function, containing a parameters required in the
        model functions, initialized with jnp.nan.

    """
    # Collect all model variables, that includes actions, states, the period, and
    # auxiliary variables (model function names).
    variables = {
        *model.functions,
        *model.actions,
        *model.states,
        "_period",
    }

    if hasattr(model, "shocks"):
        variables = variables | set(model.shocks)

    function_params = {}
    # For each model function, capture the arguments of the function that are not in the
    # set of model variables, and initialize them.
    for name, func in model.functions.items():
        arguments = set(inspect.signature(func).parameters)
        params = sorted(arguments.difference(variables))
        function_params[name] = {p: jnp.nan for p in params}

    return function_params


def _create_stochastic_transition_params(
    model: Model,
    variable_info: pd.DataFrame,
    grids: dict[str, Array],
) -> dict[str, Array]:
    """Create parameters for stochastic transitions.

    Args:
        model: The model as provided by the user.
        variable_info: A dataframe with information about the variables.
        grids: A dictionary of grids consistent with model.

    Returns:
        A dictionary of parameters required for stochastic transitions, initialized with
        jnp.nan matrices of the correct dimensions.

    """
    stochastic_variables = variable_info.query("is_stochastic").index.tolist()

    # Assert that all stochastic variables are discrete state variables
    # ==================================================================================
    discrete_state_vars = set(variable_info.query("is_state & is_discrete").index)

    invalid = set(stochastic_variables) - discrete_state_vars
    if invalid:
        raise ValueError(
            f"The following variables are stochastic, but are not discrete state "
            f"variables: {invalid}. This is currently not supported.",
        )

    # Create template matrices for stochastic transitions
    # ==================================================================================

    # Stochastic transition functions can only depend on discrete vars or '_period'.
    valid_vars = set(variable_info.query("is_discrete").index) | {"_period"}

    stochastic_transition_params = {}
    invalid_dependencies = {}

    for var in stochastic_variables:
        # Retrieve corresponding next function and its arguments
        next_var = model.functions[f"next_{var}"]
        dependencies = list(inspect.signature(next_var).parameters)

        # If there are invalid dependencies, store them in a dictionary and continue
        # with the next variable to collect as many invalid arguments as possible.
        invalid = set(dependencies) - valid_vars
        if invalid:
            invalid_dependencies[var] = invalid
        else:
            # Get the dimensions of variables that influence the stochastic variable
            dimensions_of_deps = [
                len(grids[arg]) if arg != "_period" else model.n_periods
                for arg in dependencies
            ]
            # Add the dimension of the stochastic variable itself at the end
            dimensions = (*dimensions_of_deps, len(grids[var]))

            stochastic_transition_params[var] = jnp.full(dimensions, jnp.nan)

    # Raise an error if there are invalid arguments
    # ==================================================================================
    if invalid_dependencies:
        raise ValueError(
            f"Stochastic transition functions can only depend on discrete variables or "
            "'_period'. The following variables have invalid arguments: "
            f"{invalid_dependencies}.",
        )

    return stochastic_transition_params



"""Collection of classes that are used by the user to define the model and grids."""

import dataclasses as dc
from abc import ABC, abstractmethod
from dataclasses import dataclass, fields, is_dataclass
from typing import Any

import jax.numpy as jnp
from jax import Array

from lcm import grid_helpers
from lcm.exceptions import GridInitializationError, format_messages
from lcm.typing import Scalar
from lcm.utils import find_duplicates


class Grid(ABC):
    """LCM Grid base class."""

    @abstractmethod
    def to_jax(self) -> Array:
        """Convert the grid to a Jax array."""


class DiscreteGrid(Grid):
    """A class representing a discrete grid.

    Args:
        category_class (type): The category class representing the grid categories. Must
            be a dataclass with fields that have unique scalar int or float values.

    Attributes:
        categories: The list of category names.
        codes: The list of category codes.

    Raises:
        GridInitializationError: If the `category_class` is not a dataclass with scalar
            int or float fields.

    """

    def __init__(self, category_class: type) -> None:
        """Initialize the DiscreteGrid.

        Args:
            category_class (type): The category class representing the grid categories.
                Must be a dataclass with fields that have unique scalar int or float
                values.

        """
        _validate_discrete_grid(category_class)

        names_and_values = _get_field_names_and_values(category_class)

        self.__categories = tuple(names_and_values.keys())
        self.__codes = tuple(names_and_values.values())

    @property
    def categories(self) -> tuple[str, ...]:
        """Get the list of category names."""
        return self.__categories

    @property
    def codes(self) -> tuple[int | float, ...]:
        """Get the list of category codes."""
        return self.__codes

    def to_jax(self) -> Array:
        """Convert the grid to a Jax array."""
        return jnp.array(self.codes)


@dataclass(frozen=True, kw_only=True)
class ContinuousGrid(Grid, ABC):
    """LCM Continuous Grid base class."""

    start: int | float
    stop: int | float
    n_points: int

    def __post_init__(self) -> None:
        _validate_continuous_grid(
            start=self.start,
            stop=self.stop,
            n_points=self.n_points,
        )

    @abstractmethod
    def to_jax(self) -> Array:
        """Convert the grid to a Jax array."""

    @abstractmethod
    def get_coordinate(self, value: Scalar) -> Scalar:
        """Get the generalized coordinate of a value in the grid."""

    def replace(self, **kwargs: Any) -> "ContinuousGrid":  # noqa: ANN401
        """Replace the attributes of the grid.

        Args:
            **kwargs: Keyword arguments to replace the attributes of the grid.

        Returns:
            A new grid with the replaced attributes.

        """
        try:
            return dc.replace(self, **kwargs)
        except TypeError as e:
            raise GridInitializationError(
                f"Failed to replace attributes of the grid. The error was: {e}"
            ) from e


class LinspaceGrid(ContinuousGrid):
    """A linear grid of continuous values.

    Example:
    --------
    Let `start = 1`, `stop = 100`, and `n_points = 3`. The grid is `[1, 50.5, 100]`.

    Attributes:
        start: The start value of the grid. Must be a scalar int or float value.
        stop: The stop value of the grid. Must be a scalar int or float value.
        n_points: The number of points in the grid. Must be an int greater than 0.

    """

    def to_jax(self) -> Array:
        """Convert the grid to a Jax array."""
        return grid_helpers.linspace(self.start, self.stop, self.n_points)

    def get_coordinate(self, value: Scalar) -> Scalar:
        """Get the generalized coordinate of a value in the grid."""
        return grid_helpers.get_linspace_coordinate(
            value, self.start, self.stop, self.n_points
        )


class LogspaceGrid(ContinuousGrid):
    """A logarithmic grid of continuous values.

    Example:
    --------
    Let `start = 1`, `stop = 100`, and `n_points = 3`. The grid is `[1, 10, 100]`.

    Attributes:
        start: The start value of the grid. Must be a scalar int or float value.
        stop: The stop value of the grid. Must be a scalar int or float value.
        n_points: The number of points in the grid. Must be an int greater than 0.

    """

    def to_jax(self) -> Array:
        """Convert the grid to a Jax array."""
        return grid_helpers.logspace(self.start, self.stop, self.n_points)

    def get_coordinate(self, value: Scalar) -> Scalar:
        """Get the generalized coordinate of a value in the grid."""
        return grid_helpers.get_logspace_coordinate(
            value, self.start, self.stop, self.n_points
        )


# ======================================================================================
# Validate user input
# ======================================================================================


def _validate_discrete_grid(category_class: type) -> None:
    """Validate the field names and values of the category_class passed to DiscreteGrid.

    Args:
        category_class: The category class representing the grid categories. Must
            be a dataclass with fields that have unique scalar int or float values.

    Raises:
        GridInitializationError: If the `category_class` is not a dataclass with scalar
            int or float fields.

    """
    if not is_dataclass(category_class):
        raise GridInitializationError(
            "category_class must be a dataclass with scalar int or float fields, "
            f"but is {category_class}."
        )

    names_and_values = _get_field_names_and_values(category_class)

    error_messages = []

    if not names_and_values:
        error_messages.append(
            "category_class passed to DiscreteGrid must have at least one field"
        )

    names_with_non_numerical_values = [
        name
        for name, value in names_and_values.items()
        if not isinstance(value, int | float)
    ]
    if names_with_non_numerical_values:
        error_messages.append(
            "Field values of the category_class passed to DiscreteGrid can only be "
            "scalar int or float values. The values to the following fields are not: "
            f"{names_with_non_numerical_values}"
        )

    values = list(names_and_values.values())

    duplicated_values = find_duplicates(values)
    if duplicated_values:
        error_messages.append(
            "Field values of the category_class passed to DiscreteGrid must be unique. "
            "The following values are duplicated: "
            f"{duplicated_values}"
        )

    if values != list(range(len(values))):
        error_messages.append(
            "Field values of the category_class passed to DiscreteGrid must be "
            "consecutive integers starting from 0 (e.g., 0, 1, 2, ...)."
        )

    if error_messages:
        msg = format_messages(error_messages)
        raise GridInitializationError(msg)


def _get_field_names_and_values(dc: type) -> dict[str, Any]:
    """Get the fields of a dataclass.

    Args:
        dc: The dataclass to get the fields of.

    Returns:
        A dictionary with the field names as keys and the field values as values. If
        no value is provided for a field, the value is set to None.

    """
    return {field.name: getattr(dc, field.name, None) for field in fields(dc)}


def _validate_continuous_grid(
    start: float,
    stop: float,
    n_points: int,
) -> None:
    """Validate the continuous grid parameters.

    Args:
        start: The start value of the grid.
        stop: The stop value of the grid.
        n_points: The number of points in the grid.

    Raises:
        GridInitializationError: If the grid parameters are invalid.

    """
    error_messages = []

    valid_start_type = isinstance(start, int | float)
    if not valid_start_type:
        error_messages.append("start must be a scalar int or float value")

    valid_stop_type = isinstance(stop, int | float)
    if not valid_stop_type:
        error_messages.append("stop must be a scalar int or float value")

    if not isinstance(n_points, int) or n_points < 1:
        error_messages.append(
            f"n_points must be an int greater than 0 but is {n_points}",
        )

    if valid_start_type and valid_stop_type and start >= stop:
        error_messages.append("start must be less than stop")

    if error_messages:
        msg = format_messages(error_messages)
        raise GridInitializationError(msg)



from collections.abc import Callable
from typing import cast

import jax.numpy as jnp
from dags import concatenate_functions
from dags.signature import with_signature
from jax import Array

from lcm.functools import all_as_kwargs
from lcm.grids import ContinuousGrid
from lcm.interfaces import StateSpaceInfo
from lcm.ndimage import map_coordinates
from lcm.typing import Scalar


def get_value_function_representation(
    state_space_info: StateSpaceInfo,
    *,
    input_prefix: str = "next_",
    name_of_values_on_grid: str = "next_V_arr",
) -> Callable[..., Array]:
    """Create a function representation of the next value function array.

    The returned function
    ---------------------

    This function generates a function that looks up discrete values and interpolates
    values for continuous variables on the value function array. The arguments of the
    resulting function can be split in two categories:

       1. The original arguments of the function that was used to pre-calculate the
          value function on the state space grid.

       2. Auxiliary arguments, such as information about the grids, which are needed for
          example, for the interpolation.

    After partialling in all helper arguments, the resulting function behaves like an
    analytical function, i.e. it can be evaluated on points that do not lie on the grid
    points of the state variables. In particular, it can also be jitted, differentiated
    and vmapped with jax.

    How does it work?
    -----------------

    The resulting function roughly does the following steps:

    - It translates values of discrete variables into positions.
    - It translates values of continuous variables into coordinates needed for
      interpolation via jax.scipy.ndimage.map_coordinates.
    - It performs the interpolation.

    Depending on the grid, only a subset of these steps is relevant. The chosen
    implementation of each step is also adjusted to the type of grid. In particular we
    try to avoid searching for neighboring values on a grid and instead exploit
    structure in the grid to calculate where those entries are. The order in which the
    functions are called is determined by a DAG.

    Args:
        state_space_info: Class containing all information needed to interpret the
            pre-calculated values of a function.
        input_prefix: Prefix that will be added to all argument names of the resulting
            function, except for the helpers arguments. Default is "next_"; since the
            value function is typically evaluated on the next period's state space.
        name_of_values_on_grid: The name of the argument via which the pre-calculated
            values, that have been evaluated on the state-space grid, will be passed
            into the resulting function. Defaults to "next_V_arr".

    Returns:
        A callable that lets you treat the result of pre-calculating a function on the
            state space as an analytical function.

    """
    # ==================================================================================
    # check inputs
    # ==================================================================================
    _fail_if_interpolation_axes_are_not_last(state_space_info)
    _need_interpolation = bool(state_space_info.continuous_states)

    # ==================================================================================
    # create functions to look up position of discrete variables from labels
    # ==================================================================================
    funcs = {}

    for var in state_space_info.discrete_states:
        funcs[f"__{var}_pos__"] = _get_label_translator(
            in_name=input_prefix + var,
        )

    # ==================================================================================
    # create a function for the discrete lookup
    # ==================================================================================
    # lookup is positional, so the inputs of the wrapper functions need to be the
    # outcomes of tranlating labels into positions
    _internal_axes = [f"__{var}_pos__" for var in state_space_info.states_names]
    _discrete_axes = [ax for ax in _internal_axes if ax in funcs]

    _out_name = "__interpolation_data__" if _need_interpolation else "__fval__"
    funcs[_out_name] = _get_lookup_function(
        array_name=name_of_values_on_grid,
        axis_names=_discrete_axes,
    )

    if _need_interpolation:
        # ==============================================================================
        # create functions to find coordinates for the interpolation
        # ==============================================================================
        for var, grid_spec in state_space_info.continuous_states.items():
            funcs[f"__{var}_coord__"] = _get_coordinate_finder(
                in_name=input_prefix + var,
                grid=grid_spec,
            )

        # ==============================================================================
        # create interpolation function
        # ==============================================================================
        _continuous_axes = [
            f"__{var}_coord__"
            for var in state_space_info.states_names
            if var in state_space_info.continuous_states
        ]
        funcs["__fval__"] = _get_interpolator(
            name_of_values_on_grid="__interpolation_data__",
            axis_names=_continuous_axes,
        )

    return concatenate_functions(
        functions=funcs,
        targets="__fval__",
    )


def _get_label_translator(
    in_name: str,
) -> Callable[..., Scalar]:
    """Create a function that translates a label into a position in a list of labels.

    Currently, only labels are supported that are themselves indices. The label
    translator in this case is thus just the identity function.

    Args:
        in_name: Name of the variable that provides the label in the signature of the
            resulting function.

    Returns:
        A callable with the keyword only argument `in_name` that converts a label into a
        position in a list of labels.

    """

    @with_signature(args=[in_name])
    def translate_label(*args: Scalar, **kwargs: Scalar) -> Scalar:
        kwargs = all_as_kwargs(args, kwargs, arg_names=[in_name])
        return kwargs[in_name]

    return translate_label


def _get_lookup_function(
    array_name: str,
    axis_names: list[str],
) -> Callable[..., Scalar]:
    """Create a function that emulates indexing into an array via named axes.

    Args:
        array_name (str): The name of the array into which the function indexes.
        axis_names (list): List of strings with names for each axis in the array.

    Returns:
        A callable with the keyword-only arguments `[*axis_names]` that looks up values
        from an array called `array_name`.

    """
    arg_names = [*axis_names, array_name]

    @with_signature(args=arg_names)
    def lookup_wrapper(*args: Scalar, **kwargs: Scalar) -> Scalar:
        kwargs = all_as_kwargs(args, kwargs, arg_names=arg_names)
        positions = tuple(kwargs[var] for var in axis_names)
        arr = cast(Array, kwargs[array_name])
        return arr[positions]

    return lookup_wrapper


def _get_coordinate_finder(
    in_name: str,
    grid: ContinuousGrid,
) -> Callable[..., Scalar]:
    """Create a function that translates a value into coordinates on a grid.

    The resulting coordinates can be used to do linear interpolation via
    jax.scipy.ndimage.map_coordinates.

    Args:
        in_name: Name via which the value to be translated into coordinates will be
            passed into the resulting function.
        grid: The continuous grid on which the value is to be translated into
            coordinates.

    Returns:
        A callable with keyword-only argument [in_name] that translates a value into
        coordinates on a grid.

    """

    @with_signature(args=[in_name])
    def find_coordinate(*args: Scalar, **kwargs: Scalar) -> Scalar:
        kwargs = all_as_kwargs(args, kwargs, arg_names=[in_name])
        return grid.get_coordinate(kwargs[in_name])

    return find_coordinate


def _get_interpolator(
    name_of_values_on_grid: str,
    axis_names: list[str],
) -> Callable[..., Scalar]:
    """Create a function interpolator via named axes.

    Args:
        name_of_values_on_grid: The name of the argument via which the pre-calculated
            values, that have been evaluated on a grid, will be passed into the
            resulting function.
        axis_names: Names of the axes in the data array.

    Returns:
        A callable that interpolates a function via named axes.

    """
    arg_names = [name_of_values_on_grid, *axis_names]

    @with_signature(args=arg_names)
    def interpolate(*args: Scalar, **kwargs: Scalar) -> Scalar:
        kwargs = all_as_kwargs(args, kwargs, arg_names=arg_names)
        coordinates = jnp.array([kwargs[var] for var in axis_names])
        return map_coordinates(
            input=kwargs[name_of_values_on_grid],
            coordinates=coordinates,
        )

    return interpolate


def _fail_if_interpolation_axes_are_not_last(state_space_info: StateSpaceInfo) -> None:
    """Fail if the continuous variables are not the last elements in var_names.

    Args:
        state_space_info: Class containing all information needed to interpret the
            precalculated values of a function.

    Raises:
        ValueError: If the continuous variables are not the last elements in var_names.

    """
    common = set(state_space_info.continuous_states) & set(
        state_space_info.states_names
    )

    if common:
        n_common = len(common)
        if sorted(common) != sorted(state_space_info.states_names[-n_common:]):
            msg = "Continuous variables need to be the last entries in var_names."
            raise ValueError(msg)



from collections.abc import Callable
from functools import partial
from typing import Literal

import jax
import pandas as pd
from jax import Array

from lcm.input_processing import process_model
from lcm.interfaces import StateActionSpace, StateSpaceInfo
from lcm.logging import get_logger
from lcm.max_Q_over_c import (
    get_argmax_and_max_Q_over_c,
    get_max_Q_over_c,
)
from lcm.max_Qc_over_d import get_max_Qc_over_d
from lcm.next_state import get_next_state_function
from lcm.Q_and_F import (
    get_Q_and_F,
)
from lcm.simulation.simulate import simulate, solve_and_simulate
from lcm.solution.solve_brute import solve
from lcm.state_action_space import (
    create_state_action_space,
    create_state_space_info,
)
from lcm.typing import (
    ArgmaxQOverCFunction,
    MaxQcOverDFunction,
    MaxQOverCFunction,
    ParamsDict,
    Target,
)
from lcm.user_model import Model


def get_lcm_function(
    model: Model,
    *,
    targets: Literal["solve", "simulate", "solve_and_simulate"],
    debug_mode: bool = True,
    jit: bool = True,
) -> tuple[Callable[..., dict[int, Array] | pd.DataFrame], ParamsDict]:
    """Entry point for users to get high level functions generated by lcm.

    Return the function to solve and/or simulate a model along with a template for the
    parameters.

    Advanced users might want to use lower level functions instead, but can read the
    source code of this function to see how the lower level components are meant to be
    used.

    Args:
        model: User model specification.
        targets: The requested function types. Currently only "solve", "simulate" and
            "solve_and_simulate" are supported.
        debug_mode: Whether to log debug messages.
        jit: Whether to jit the returned function.

    Returns:
        - A function that takes params (and possibly other arguments, such as initial
          states in the simulate case) and returns the requested targets.
        - A parameter dictionary where all parameter values are initialized to NaN.

    """
    # ==================================================================================
    # preparations
    # ==================================================================================
    if targets not in {"solve", "simulate", "solve_and_simulate"}:
        raise NotImplementedError

    internal_model = process_model(model)
    last_period = internal_model.n_periods - 1

    logger = get_logger(debug_mode=debug_mode)

    # ==================================================================================
    # Create model functions and state-action-spaces
    # ==================================================================================
    state_action_spaces: dict[int, StateActionSpace] = {}
    state_space_infos: dict[int, StateSpaceInfo] = {}
    max_Q_over_c_functions: dict[int, MaxQOverCFunction] = {}
    argmax_and_max_Q_over_c_functions: dict[int, ArgmaxQOverCFunction] = {}
    max_Qc_over_d_functions: dict[int, MaxQcOverDFunction] = {}

    for period in reversed(range(internal_model.n_periods)):
        is_last_period = period == last_period

        state_action_space = create_state_action_space(
            model=internal_model,
            is_last_period=is_last_period,
        )

        state_space_info = create_state_space_info(
            model=internal_model,
            is_last_period=is_last_period,
        )

        if is_last_period:
            next_state_space_info = LastPeriodsNextStateSpaceInfo
        else:
            next_state_space_info = state_space_infos[period + 1]

        Q_and_F = get_Q_and_F(
            model=internal_model,
            next_state_space_info=next_state_space_info,
            period=period,
        )

        max_Q_over_c = get_max_Q_over_c(
            Q_and_F=Q_and_F,
            continuous_actions_names=tuple(state_action_space.continuous_actions),
            states_and_discrete_actions_names=state_action_space.states_and_discrete_actions_names,
        )

        argmax_and_max_Q_over_c = get_argmax_and_max_Q_over_c(
            Q_and_F=Q_and_F,
            continuous_actions_names=tuple(state_action_space.continuous_actions),
        )

        max_Qc_over_d = get_max_Qc_over_d(
            random_utility_shock_type=internal_model.random_utility_shocks,
            variable_info=internal_model.variable_info,
            is_last_period=is_last_period,
        )

        state_action_spaces[period] = state_action_space
        state_space_infos[period] = state_space_info
        max_Q_over_c_functions[period] = max_Q_over_c
        argmax_and_max_Q_over_c_functions[period] = argmax_and_max_Q_over_c
        max_Qc_over_d_functions[period] = max_Qc_over_d

    # ==================================================================================
    # select requested solver and partial arguments into it
    # ==================================================================================
    _solve_model = partial(
        solve,
        state_action_spaces=state_action_spaces,
        max_Q_over_c_functions=max_Q_over_c_functions,
        max_Qc_over_d_functions=max_Qc_over_d_functions,
        logger=logger,
    )
    solve_model = jax.jit(_solve_model) if jit else _solve_model

    _next_state_simulate = get_next_state_function(
        model=internal_model, target=Target.SIMULATE
    )
    next_state_simulate = jax.jit(_next_state_simulate) if jit else _next_state_simulate
    simulate_model = partial(
        simulate,
        argmax_and_max_Q_over_c_functions=argmax_and_max_Q_over_c_functions,
        model=internal_model,
        next_state=next_state_simulate,  # type: ignore[arg-type]
        logger=logger,
    )

    solve_and_simulate_model = partial(
        solve_and_simulate,
        argmax_and_max_Q_over_c_functions=argmax_and_max_Q_over_c_functions,
        model=internal_model,
        next_state=next_state_simulate,  # type: ignore[arg-type]
        logger=logger,
        solve_model=solve_model,
    )

    target_func: Callable[..., dict[int, Array] | pd.DataFrame]

    if targets == "solve":
        target_func = solve_model
    elif targets == "simulate":
        target_func = simulate_model
    elif targets == "solve_and_simulate":
        target_func = solve_and_simulate_model

    return target_func, internal_model.params


LastPeriodsNextStateSpaceInfo = StateSpaceInfo(
    states_names=(),
    discrete_states={},
    continuous_states={},
)



import os
from functools import partial

import jax
from jax import Array


def random_choice(
    labels: jax.Array,
    probs: jax.Array,
    key: jax.Array,
) -> jax.Array:
    """Draw multiple random choices.

    Args:
        labels: 1d array of labels.
        probs: 2d array of probabilities. Second dimension must be the same length as
            the dimension of labels.
        key: Random key.

    Returns:
        Selected labels. 1d array of length len(probs).

    """
    keys = jax.random.split(key, probs.shape[0])
    return _vmapped_choice(keys, probs, labels)


@partial(jax.vmap, in_axes=(0, 0, None))
def _vmapped_choice(key: jax.Array, probs: jax.Array, labels: jax.Array) -> jax.Array:
    return jax.random.choice(key, a=labels, p=probs)


def generate_simulation_keys(
    key: Array, ids: list[str]
) -> tuple[Array, dict[str, Array]]:
    """Generate pseudo-random number generator keys (PRNG keys) for simulation.

    PRNG keys in JAX are immutable objects used to control random number generation.
    A key can be used to generate a stream of random numbers, e.g., given a key, one can
    call jax.random.normal(key) to generate a stream of normal random numbers. In order
    to ensure that each simulation is based on a different stream of random numbers, we
    split the key into one key per stochastic variable, and one key that will be passed
    to the next iteration in order to generate new keys.

    See the JAX documentation for more details:
    https://docs.jax.dev/en/latest/random-numbers.html#random-numbers-in-jax

    Args:
        key: Random key.
        ids: List of names for which a key is to be generated.

    Returns:
        - Updated random key.
        - Dict with random keys for each id in ids.

    """
    keys = jax.random.split(key, num=len(ids) + 1)

    next_key = keys[0]
    simulation_keys = dict(zip(ids, keys[1:], strict=True))

    return next_key, simulation_keys


def draw_random_seed() -> int:
    """Generate a random seed using the operating system's secure entropy pool.

    Returns:
        Random seed.

    """
    return int.from_bytes(os.urandom(4), "little")



import dataclasses
from collections.abc import Mapping

import pandas as pd
from jax import Array

from lcm.grids import ContinuousGrid, DiscreteGrid, Grid
from lcm.typing import InternalUserFunction, ParamsDict, ShockType
from lcm.utils import first_non_none


@dataclasses.dataclass(frozen=True)
class StateActionSpace:
    """The state-action space.

    When used for the model solution:
    ---------------------------------

    The state-action space becomes the full Cartesian product of the state variables and
    the action variables.

    When used for the simulation:
    ----------------------------

    The state-action space becomes the product of state-combinations with the full
    Cartesian product of the action variables.

    In both cases, infeasible state-action combinations will be masked.

    Note:
    -----
    We store discrete and continuous actions separately since these are handled during
    different stages of the solution and simulation processes.

    Attributes:
        states: Dictionary containing the values of the state variables.
        discrete_actions: Dictionary containing the values of the discrete action
            variables.
        continuous_actions: Dictionary containing the values of the continuous action
            variables.
        states_and_discrete_actions_names: Tuple with names of states and discrete
            action variables in the order they appear in the variable info table.

    """

    states: dict[str, Array]
    discrete_actions: dict[str, Array]
    continuous_actions: dict[str, Array]
    states_and_discrete_actions_names: tuple[str, ...]

    def replace(
        self,
        states: dict[str, Array] | None = None,
        discrete_actions: dict[str, Array] | None = None,
        continuous_actions: dict[str, Array] | None = None,
    ) -> "StateActionSpace":
        """Replace the states or actions in the state-action space.

        Args:
            states: Dictionary with new states. If None, the existing states are used.
            discrete_actions: Dictionary with new discrete actions. If None, the
                existing discrete actions are used.
            continuous_actions: Dictionary with new continuous actions. If None, the
                existing continuous actions are used.

        Returns:
            New state-action space with the replaced states or actions.

        """
        states = first_non_none(states, self.states)
        discrete_actions = first_non_none(discrete_actions, self.discrete_actions)
        continuous_actions = first_non_none(continuous_actions, self.continuous_actions)
        return dataclasses.replace(
            self,
            states=states,
            discrete_actions=discrete_actions,
            continuous_actions=continuous_actions,
        )


@dataclasses.dataclass(frozen=True)
class StateSpaceInfo:
    """Information to work with the output of a function evaluated on a state space.

    An example is the value function array, which is the output of the value function
    evaluated on the state space.

    Attributes:
        var_names: Tuple with names of state variables.
        discrete_vars: Dictionary with grids of discrete state variables.
        continuous_vars: Dictionary with grids of continuous state variables.

    """

    states_names: tuple[str, ...]
    discrete_states: Mapping[str, DiscreteGrid]
    continuous_states: Mapping[str, ContinuousGrid]


@dataclasses.dataclass(frozen=True)
class InternalModel:
    """Internal representation of a user model.

    Attributes:
        grids: Dictionary that maps names of model variables to grids of feasible values
            for that variable.
        gridspecs: Dictionary that maps names of model variables to specifications from
            which grids of feasible values can be built.
        variable_info: A table with information about all variables in the model. The
            index contains the name of a model variable. The columns are booleans that
            are True if the variable has the corresponding property. The columns are:
            is_state, is_action, is_continuous, is_discrete.
        functions: Dictionary that maps names of functions to functions. The functions
            differ from the user functions in that they take `params` as a keyword
            argument. Two cases:
            - If the original function depended on model parameters, those are
              automatically extracted from `params` and passed to the original
              function.
            - Otherwise, the `params` argument is simply ignored.
        function_info: A table with information about all functions in the model. The
            index contains the name of a function. The columns are booleans that are
            True if the function has the corresponding property. The columns are:
            is_constraint, is_next.
        params: Dict of model parameters.
        n_periods: Number of periods.
        random_utility_shocks: Type of random utility shocks.

    """

    grids: dict[str, Array]
    gridspecs: dict[str, Grid]
    variable_info: pd.DataFrame
    functions: dict[str, InternalUserFunction]
    function_info: pd.DataFrame
    params: ParamsDict
    n_periods: int
    # Not properly processed yet
    random_utility_shocks: ShockType


@dataclasses.dataclass(frozen=True)
class InternalSimulationPeriodResults:
    """The results of a simulation for one period."""

    value: Array
    actions: dict[str, Array]
    states: dict[str, Array]



from functools import partial

import jax
import pandas as pd
from jax import Array

from lcm.argmax import argmax_and_max
from lcm.typing import (
    ArgmaxQcOverDFunction,
    MaxQcOverDFunction,
    ParamsDict,
    ShockType,
)


def get_max_Qc_over_d(
    *,
    random_utility_shock_type: ShockType,
    variable_info: pd.DataFrame,
    is_last_period: bool,
) -> MaxQcOverDFunction:
    r"""Get the function returning the maximum of Qc over discrete actions.

    The state-action value function $Q$ is defined as:

    ```{math}
    Q(x, a) =  H(U(x, a), \mathbb{E}[V(x', a') | x, a]),
    ```
    with $H(U, v) = u + \beta \cdot v$ as the leading case (which is the only one that
    is pre-implemented in LCM).

    Fixing a state and discrete action, maximizing over the feasible continuous actions,
    we get the $Q^c$ function:

    ```{math}
    Q^{c}(x, a^d) = \max_{a^c} Q(x, a^d, a^c).
    ```

    And maximizing over the discrete actions, we get the value function:

    ```{math}
    V(x) = \max_{a^d} Q^{c}(x, a^d).
    ```

    This last step is handled by the function returned here.

    Args:
        random_utility_shock_type: Type of action shock. Currently only Shock.NONE is
            supported. Work for "extreme_value" is in progress.
        variable_info: DataFrame with information about the variables.
        is_last_period: Whether the function is created for the last period.

    Returns:
        Function that returns the argument that maximize the Qc-function over the
        discrete actions. The maximizing argument corresponds to the policy function of
        the discrete actions.

    """
    if is_last_period:
        variable_info = variable_info.query("~is_auxiliary")

    discrete_action_axes = _determine_discrete_action_axes_solution(variable_info)

    if random_utility_shock_type == ShockType.NONE:
        func = _max_Qc_over_d_no_shocks
    elif random_utility_shock_type == ShockType.EXTREME_VALUE:
        raise NotImplementedError("Extreme value shocks are not yet implemented.")
    else:
        raise ValueError(f"Invalid shock_type: {random_utility_shock_type}.")

    return partial(func, discrete_action_axes=discrete_action_axes)


def get_argmax_and_max_Qc_over_d(
    *,
    variable_info: pd.DataFrame,
) -> ArgmaxQcOverDFunction:
    r"""Get the function returning the arguments maximizing Qc over discrete actions.

    The state-action value function $Q$ is defined as:

    ```{math}
    Q(x, a) =  H(U(x, a), \mathbb{E}[V(x', a') | x, a]),
    ```
    with $H(U, v) = u + \beta \cdot v$ as the leading case (which is the only one that
    is pre-implemented in LCM).

    Fixing a state and discrete action, maximizing over the feasible continuous actions,
    we get the $Q^c$ function:

    ```{math}
    Q^{c}(x, a^d) = \max_{a^c} Q(x, a^d, a^c).
    ```

    Taking the argmax over the discrete actions, we get the policy function of the
    discrete actions:

    ```{math}
    \pi^{d}(x) = \argmax_{a^d} Q^{c}(x, a^d).
    ```

    This last step is handled by the function returned here.

    Args:
        variable_info: DataFrame with information about the variables.

    Returns:
        Function that returns the arguments that maximize the Qc-function over the
        discrete actions and the maximum itself, i.e., policy function of the discrete
        actions. The maximum corresponds to the value function.

    """
    discrete_action_axes = _determine_discrete_action_axes_simulation(variable_info)

    def argmax_and_max_Qc_over_d(
        Qc_arr: Array,
        discrete_action_axes: tuple[int, ...],
        params: ParamsDict,  # noqa: ARG001
    ) -> tuple[Array, Array]:
        return argmax_and_max(Qc_arr, axis=discrete_action_axes)

    return partial(argmax_and_max_Qc_over_d, discrete_action_axes=discrete_action_axes)


# ======================================================================================
# Discrete problem with no shocks
# ======================================================================================


def _max_Qc_over_d_no_shocks(
    Qc_arr: Array,
    discrete_action_axes: tuple[int, ...],
    params: ParamsDict,  # noqa: ARG001
) -> Array:
    """Take the maximum of the Qc-function over the discrete actions.

    Args:
        Qc_arr: The maximum of the state-action value function (Q) over the continuous
            actions, conditional on the discrete action. This has one axis for each
            state and discrete action variable.
        discrete_action_axes: Tuple of indices representing the axes in the value
            function that correspond to discrete actions.
        params: See `get_solve_discrete_problem`.

    Returns:
        The maximum of Qc_arr over the discrete action axes.

    """
    return Qc_arr.max(axis=discrete_action_axes)


# ======================================================================================
# Discrete problem with extreme value shocks
# --------------------------------------------------------------------------------------
# The following is currently *NOT* supported.
# ======================================================================================


def _max_Qc_over_d_extreme_value_shocks(
    Qc_arr: Array, discrete_action_axes: tuple[int, ...], params: ParamsDict
) -> Array:
    """Take the expected maximum of the Qc-function over the discrete actions.

    Args:
        Qc_arr: The maximum of the state-action value function (Q) over the continuous
            actions, conditional on the discrete action. This has one axis for each
            state and discrete action variable.
        discrete_action_axes: Tuple of indices representing the axes in the value
            function that correspond to discrete actions.
        params: See `get_solve_discrete_problem`.

    Returns:
        The expected maximum of Qc_arr over the discrete action axes.

    """
    scale = params["additive_utility_shock"]["scale"]
    return scale * jax.scipy.special.logsumexp(
        Qc_arr / scale, axis=discrete_action_axes
    )


# ======================================================================================
# Auxiliary functions
# ======================================================================================


def _determine_discrete_action_axes_solution(
    variable_info: pd.DataFrame,
) -> tuple[int, ...]:
    """Get axes of state-action-space that correspond to discrete actions in solution.

    Args:
        variable_info: DataFrame with information about the variables.

    Returns:
        A tuple of indices representing the axes' positions in the value function that
        correspond to discrete actions.

    """
    discrete_action_vars = set(
        variable_info.query("is_action & is_discrete").index.tolist()
    )
    return tuple(
        i for i, ax in enumerate(variable_info.index) if ax in discrete_action_vars
    )


def _determine_discrete_action_axes_simulation(
    variable_info: pd.DataFrame,
) -> tuple[int, ...]:
    """Get axes of state-action-space that correspond to discrete actions in simulation.

    Args:
        variable_info: DataFrame with information about the variables.

    Returns:
        A tuple of indices representing the axes' positions in the value function that
        correspond to discrete actions.

    """
    discrete_action_vars = set(
        variable_info.query("is_action & is_discrete").index.tolist()
    )

    # The first dimension corresponds to the simulated states, so add 1.
    return tuple(1 + i for i in range(len(discrete_action_vars)))



"""Functions to generate and work with different kinds of grids.

Grid generation functions must have the following signature:

    Signature (start: Scalar, stop: Scalar, n_points: int) -> jax.Array

They take start and end points and create a grid of points between them.


Interpolation info functions must have the following signature:

    Signature (
        value: Scalar,
        start: Scalar,
        stop: Scalar,
        n_points: int
    ) -> Scalar

They take the information required to generate a grid, and return an index corresponding
to the value, which is a point in the space but not necessarily a grid point.

Some of the arguments will not be used by all functions but the aligned interface makes
it easy to call functions interchangeably.

"""

import jax.numpy as jnp
from jax import Array

from lcm.typing import Scalar


def linspace(start: Scalar, stop: Scalar, n_points: int) -> Array:
    """Wrapper around jnp.linspace.

    Returns a linearly spaced grid between start and stop with n_points, including both
    endpoints.

    """
    return jnp.linspace(start, stop, n_points)


def get_linspace_coordinate(
    value: Scalar,
    start: Scalar,
    stop: Scalar,
    n_points: int,
) -> Scalar:
    """Map a value into the input needed for jax.scipy.ndimage.map_coordinates."""
    step_length = (stop - start) / (n_points - 1)
    return (value - start) / step_length


def logspace(start: Scalar, stop: Scalar, n_points: int) -> Array:
    """Wrapper around jnp.logspace.

    Returns a logarithmically spaced grid between start and stop with n_points,
    including both endpoints.

    From the JAX documentation:

        In linear space, the sequence starts at base ** start (base to the power of
        start) and ends with base ** stop [...].

    """
    start_linear = jnp.log(start)
    stop_linear = jnp.log(stop)
    return jnp.logspace(start_linear, stop_linear, n_points, base=jnp.e)


def get_logspace_coordinate(
    value: Scalar,
    start: Scalar,
    stop: Scalar,
    n_points: int,
) -> Scalar:
    """Map a value into the input needed for jax.scipy.ndimage.map_coordinates."""
    # Transform start, stop, and value to linear scale
    start_linear = jnp.log(start)
    stop_linear = jnp.log(stop)
    value_linear = jnp.log(value)

    # Calculate coordinate in linear space
    coordinate_in_linear_space = get_linspace_coordinate(
        value_linear,
        start_linear,
        stop_linear,
        n_points,
    )

    # Calculate rank of lower and upper point in logarithmic space
    rank_lower_gridpoint = jnp.floor(coordinate_in_linear_space)
    rank_upper_gridpoint = rank_lower_gridpoint + 1

    # Calculate lower and upper point in logarithmic space
    step_length_linear = (stop_linear - start_linear) / (n_points - 1)
    lower_gridpoint = jnp.exp(start_linear + step_length_linear * rank_lower_gridpoint)
    upper_gridpoint = jnp.exp(start_linear + step_length_linear * rank_upper_gridpoint)

    # Calculate the decimal part of coordinate
    logarithmic_step_size_at_coordinate = upper_gridpoint - lower_gridpoint
    distance_from_lower_gridpoint = value - lower_gridpoint

    # If the distance from the lower gridpoint is zero, the coordinate corresponds to
    # the rank of the lower gridpoint. The other extreme is when the distance is equal
    # to the logarithmic step size at the coordinate, in which case the coordinate
    # corresponds to the rank of the upper gridpoint. For values in between, the
    # coordinate lies on a linear scale between the ranks of the lower and upper
    # gridpoints.
    decimal_part = distance_from_lower_gridpoint / logarithmic_step_size_at_coordinate
    return rank_lower_gridpoint + decimal_part



import inspect
from collections.abc import Callable
from typing import Literal, TypeVar, cast

from jax import Array, vmap

from lcm.functools import allow_args, allow_only_kwargs
from lcm.utils import find_duplicates

FunctionWithArrayReturn = TypeVar(
    "FunctionWithArrayReturn", bound=Callable[..., Array | tuple[Array, Array]]
)


def simulation_spacemap(
    func: FunctionWithArrayReturn,
    actions_names: tuple[str, ...],
    states_names: tuple[str, ...],
) -> FunctionWithArrayReturn:
    """Apply vmap such that func can be evaluated on actions and simulation states.

    This function maps the function `func` over the simulation state-action-space. That
    is, it maps `func` over the Cartesian product of the action variables, and over the
    fixed simulation states. For each action variable, a leading dimension is added to
    the output object, with the length of the axis being the number of possible values
    in the grid. Importantly, it does not create a Cartesian product over the state
    variables, since these are fixed during the simulation. For the state variables,
    a single dimension is added to the output object, with the length of the axis
    being the number of simulated states.

    simulation_spacemap preserves the function signature and allows the function to be
    called with keyword arguments.

    Args:
        func: The function to be dispatched.
        actions_names: Names of the action variables.
        states_names: Names of the state variables.

    Returns:
        A callable with the same arguments as func (but with an additional leading
        dimension) that returns an Array or pytree of Arrays. If `func` returns a
        scalar, the dispatched function returns an Array with k + 1 dimensions, where k
        is the length of `actions_names` and the additional dimension corresponds to the
        `states_names`. The order of the dimensions is determined by the order of
        `actions_names`. If the output of `func` is a jax pytree, the usual jax behavior
        applies, i.e. the leading dimensions of all arrays in the pytree are as
        described above but there might be additional dimensions.

    """
    if duplicates := find_duplicates(actions_names, states_names):
        msg = (
            "Same argument provided more than once in actions or states variables, "
            f"or is present in both: {duplicates}"
        )
        raise ValueError(msg)

    mappable_func = allow_args(func)

    vmapped = _base_productmap(mappable_func, actions_names)
    vmapped = vmap_1d(vmapped, variables=states_names, callable_with="only_args")

    # This raises a mypy error but is perfectly fine to do. See
    # https://github.com/python/mypy/issues/12472
    vmapped.__signature__ = inspect.signature(mappable_func)  # type: ignore[attr-defined]

    return cast(FunctionWithArrayReturn, allow_only_kwargs(vmapped))


def vmap_1d(
    func: FunctionWithArrayReturn,
    variables: tuple[str, ...],
    *,
    callable_with: Literal["only_args", "only_kwargs"] = "only_kwargs",
) -> FunctionWithArrayReturn:
    """Apply vmap such that func is mapped over the specified variables.

    In contrast to a general vmap call, vmap_1d vectorizes along the leading axis of all
    of the requested variables simultaneously. Moreover, it preserves the function
    signature and allows the function to be called with keyword arguments.

    Args:
        func: The function to be dispatched.
        variables: Tuple with names of arguments that over which we map.
        callable_with: Whether to apply the allow_kwargs decorator to the dispatched
            function. If "only_args", the returned function can only be called with
            positional arguments. If "only_kwargs", the returned function can only be
            called with keyword arguments.

    Returns:
        A callable with the same arguments as func (but with an additional leading
        dimension) that returns a jax.Array or pytree of arrays. If `func`
        returns a scalar, the dispatched function returns a jax.Array with 1
        jax.Array with 1 dimension and length k, where k is the length of one of
        the mapped inputs in `variables`. The order of the dimensions is determined by
        the order of `variables` which can be different to the order of `funcs`
        arguments. If the output of `func` is a jax pytree, the usual jax behavior
        applies, i.e. the leading dimensions of all arrays in the pytree are as
        described above but there might be additional dimensions.

    """
    if duplicates := find_duplicates(variables):
        raise ValueError(
            f"Same argument provided more than once in variables: {duplicates}",
        )

    signature = inspect.signature(func)
    parameters = list(signature.parameters)

    positions = [parameters.index(var) for var in variables]

    # Create in_axes to apply vmap over variables. This has one entry for each argument
    # of func, indicating whether the argument should be mapped over or not. None means
    # that the argument should not be mapped over, 0 means that it should be mapped over
    # the leading axis of the input.
    in_axes_for_vmap = [None] * len(parameters)  # type: list[int | None]
    for p in positions:
        in_axes_for_vmap[p] = 0

    vmapped = vmap(func, in_axes=in_axes_for_vmap)

    # This raises a mypy error but is perfectly fine to do. See
    # https://github.com/python/mypy/issues/12472
    vmapped.__signature__ = signature  # type: ignore[attr-defined]

    if callable_with == "only_kwargs":
        out = allow_only_kwargs(vmapped)
    elif callable_with == "only_args":
        out = vmapped
    else:
        raise ValueError(
            f"Invalid callable_with option: {callable_with}. Possible options are "
            "('only_args', 'only_kwargs')",
        )

    return cast(FunctionWithArrayReturn, out)


def productmap(
    func: FunctionWithArrayReturn, variables: tuple[str, ...]
) -> FunctionWithArrayReturn:
    """Apply vmap such that func is evaluated on the Cartesian product of variables.

    This is achieved by an iterative application of vmap.

    In contrast to _base_productmap, productmap preserves the function signature and
    allows the function to be called with keyword arguments.

    Args:
        func: The function to be dispatched.
        variables: Tuple with names of arguments that over which the Cartesian product
            should be formed.

    Returns:
        A callable with the same arguments as func (but with an additional leading
        dimension) that returns a jax.Array or pytree of arrays. If `func`
        returns a scalar, the dispatched function returns a jax.Array with k
        dimensions, where k is the length of `variables`. The order of the dimensions
        is determined by the order of `variables` which can be different to the order
        of `funcs` arguments. If the output of `func` is a jax pytree, the usual jax
        behavior applies, i.e. the leading dimensions of all arrays in the pytree are as
        described above but there might be additional dimensions.

    """
    if duplicates := find_duplicates(variables):
        raise ValueError(
            f"Same argument provided more than once in variables: {duplicates}",
        )

    func_callable_with_args = allow_args(func)

    vmapped = _base_productmap(func_callable_with_args, variables)

    # This raises a mypy error but is perfectly fine to do. See
    # https://github.com/python/mypy/issues/12472
    vmapped.__signature__ = inspect.signature(func_callable_with_args)  # type: ignore[attr-defined]

    return cast(FunctionWithArrayReturn, allow_only_kwargs(vmapped))


def _base_productmap(
    func: FunctionWithArrayReturn, product_axes: tuple[str, ...]
) -> FunctionWithArrayReturn:
    """Map func over the Cartesian product of product_axes.

    Like vmap, this function does not preserve the function signature and does not allow
    the function to be called with keyword arguments.

    Args:
        func: The function to be dispatched. Cannot have keyword-only arguments.
        product_axes: Tuple with names of arguments over which we apply vmap.

    Returns:
        A callable with the same arguments as func. See `product_map` for details.

    """
    signature = inspect.signature(func)
    parameters = list(signature.parameters)

    positions = [parameters.index(ax) for ax in product_axes]

    vmap_specs = []
    # We iterate in reverse order such that the output dimensions are in the same order
    # as the input dimensions.
    for pos in reversed(positions):
        spec: list[int | None] = [None] * len(parameters)
        spec[pos] = 0
        vmap_specs.append(spec)

    vmapped = func
    for spec in vmap_specs:
        vmapped = vmap(vmapped, in_axes=spec)

    return vmapped



import functools
from collections.abc import Callable

import jax.numpy as jnp
from jax import Array

from lcm.argmax import argmax_and_max
from lcm.dispatchers import productmap
from lcm.typing import ArgmaxQOverCFunction, MaxQOverCFunction, ParamsDict, Scalar


def get_max_Q_over_c(
    Q_and_F: Callable[..., tuple[Array, Array]],
    continuous_actions_names: tuple[str, ...],
    states_and_discrete_actions_names: tuple[str, ...],
) -> MaxQOverCFunction:
    r"""Get the function returning the maximum of Q over continuous actions.

    The state-action value function $Q$ is defined as:

    ```{math}
    Q(x, a) =  H(U(x, a), \mathbb{E}[V(x', a') | x, a]),
    ```
    with $H(U, v) = u + \beta \cdot v$ as the leading case (which is the only one that
    is pre-implemented in LCM).

    Fixing a state and discrete action, maximizing over the feasible continuous actions,
    we get the $Q^c$ function:

    ```{math}
    Q^{c}(x, a^d) = \max_{a^c} Q(x, a^d, a^c).
    ```

    This last step is handled by the function returned here.

    Args:
        Q_and_F: A function that takes a state-action combination and returns the action
            value of that combination and whether the state-action combination is
            feasible.
        continuous_actions_names: Tuple of action variable names that are continuous.
        states_and_discrete_actions_names: Tuple of state and discrete action variable
            names.

    Returns:
        Qc, i.e., the function that calculates the maximum of the Q-function over the
        feasible continuous actions.

    """
    if continuous_actions_names:
        Q_and_F = productmap(
            func=Q_and_F,
            variables=continuous_actions_names,
        )

    @functools.wraps(Q_and_F)
    def max_Q_over_c(next_V_arr: Array, params: ParamsDict, **kwargs: Scalar) -> Array:
        Q_arr, F_arr = Q_and_F(params=params, next_V_arr=next_V_arr, **kwargs)
        return Q_arr.max(where=F_arr, initial=-jnp.inf)

    return productmap(max_Q_over_c, variables=states_and_discrete_actions_names)


def get_argmax_and_max_Q_over_c(
    Q_and_F: Callable[..., tuple[Array, Array]],
    continuous_actions_names: tuple[str, ...],
) -> ArgmaxQOverCFunction:
    r"""Get the function returning the arguments maximizing Q over continuous actions.

    The state-action value function $Q$ is defined as:

    ```{math}
    Q(x, a) =  H(U(x, a), \mathbb{E}[V(x', a') | x, a]),
    ```
    with $H(U, v) = u + \beta \cdot v$ as the leading case (which is the only one that
    is pre-implemented in LCM).

    Fixing a state and discrete action but choosing the feasible continuous actions that
    maximizes Q, we get

    ```{math}
    \pi^{c}(x, a^d) = \argmax_{a^c} Q(x, a^d, a^c).
    ```

    This last step is handled by the function returned here.

    Args:
        Q_and_F: A function that takes a state-action combination and returns the action
            value of that combination and whether the state-action combination is
            feasible.
        continuous_actions_names: Tuple of action variable names that are continuous.

    Returns:
        Function that calculates the argument maximizing Q over the feasible continuous
        actions and the maximum iteself. The argument maximizing Q is the policy
        function of the continuous actions, conditional on the states and discrete
        actions. The maximum corresponds to the Qc-function.

    """
    if continuous_actions_names:
        Q_and_F = productmap(
            func=Q_and_F,
            variables=continuous_actions_names,
        )

    @functools.wraps(Q_and_F)
    def argmax_and_max_Q_over_c(
        next_V_arr: Array, params: ParamsDict, **kwargs: Scalar
    ) -> tuple[Array, Array]:
        Q_arr, F_arr = Q_and_F(params=params, next_V_arr=next_V_arr, **kwargs)
        return argmax_and_max(Q_arr, where=F_arr, initial=-jnp.inf)

    return argmax_and_max_Q_over_c



class ModelInitilizationError(Exception):
    """Raised when there is an error in the model initialization."""


class GridInitializationError(Exception):
    """Raised when there is an error in the grid initialization."""


def format_messages(errors: str | list[str]) -> str:
    """Convert message or list of messages into a single string."""
    if isinstance(errors, str):
        formatted = errors
    elif len(errors) == 1:
        formatted = errors[0]
    else:
        enumerated = "\n\n".join([f"{i}. {error}" for i, error in enumerate(errors, 1)])
        formatted = f"The following errors occurred:\n\n{enumerated}"
    return formatted



import jax.numpy as jnp
from jax import Array

# ======================================================================================
# argmax
# ======================================================================================


def argmax_and_max(
    a: Array,
    axis: int | tuple[int, ...] | None = None,
    initial: float | None = None,
    where: Array | None = None,
) -> tuple[Array, Array]:
    """Compute the argmax of an n-dim array along axis.

    If multiple maxima exist, the first index will be selected.

    Args:
        a: Multidimensional array.
        axis: Axis along which to compute the argmax. If None, the argmax is computed
            over all axes.
        initial: The minimum value of an output element. Must be present to
            allow computation on empty slice. See ~numpy.ufunc.reduce for details.
        where: Elements to compare for the maximum. See ~numpy.ufunc.reduce
            for details.

    Returns:
        - The argmax indices. Array with the same shape as a, except for the dimensions
          specified in axis, which are dropped. The value corresponds to an index that
          can be translated into a tuple of indices using jnp.unravel_index.
        - The corresponding maximum values.

    """
    # Preparation
    # ==================================================================================
    if axis is None:
        axis = tuple(range(a.ndim))
    elif isinstance(axis, int):
        axis = (axis,)

    # Move axis over which to compute the argmax to the back and flatten last dims
    # ==================================================================================
    a = _move_axes_to_back(a, axes=axis)
    a = _flatten_last_n_axes(a, n=len(axis))

    # Do same transformation for where
    # ==================================================================================
    if where is not None:
        where = _move_axes_to_back(where, axes=axis)
        where = _flatten_last_n_axes(where, n=len(axis))

    # Compute argmax over last dimension
    # ----------------------------------------------------------------------------------
    # Note: If multiple maxima exist, this approach will select the first index.
    # ==================================================================================
    _max = jnp.max(a, axis=-1, keepdims=True, initial=initial, where=where)
    max_value_mask = a == _max
    if where is not None:
        max_value_mask = jnp.logical_and(max_value_mask, where)
    _argmax = jnp.argmax(max_value_mask, axis=-1)

    return _argmax, _max.reshape(_argmax.shape)


def _move_axes_to_back(a: Array, axes: tuple[int, ...]) -> Array:
    """Move specified axes to the back of the array.

    Args:
        a: Multidimensional jax array.
        axes: Axes to move to the back.

    Returns:
        Array a with shifted axes.

    """
    front_axes = sorted(set(range(a.ndim)) - set(axes))
    return a.transpose((*front_axes, *axes))


def _flatten_last_n_axes(a: Array, n: int) -> Array:
    """Flatten the last n axes of a to 1 dimension.

    Args:
        a: Multidimensional jax array.
        n: Number of axes to flatten.

    Returns:
        Array a with flattened last n axes.

    """
    return a.reshape(*a.shape[:-n], -1)



"""Generate function that compute the next states for solution and simulation."""

from collections.abc import Callable

from dags import concatenate_functions
from dags.signature import with_signature
from jax import Array

from lcm.interfaces import InternalModel
from lcm.random import random_choice
from lcm.typing import Scalar, StochasticNextFunction, Target


def get_next_state_function(
    model: InternalModel,
    target: Target,
) -> Callable[..., dict[str, Scalar]]:
    """Get function that computes the next states during the solution.

    Args:
        model: Internal model instance.
        target: Whether to generate the function for the solve or simulate target.

    Returns:
        Function that computes the next states. Depends on states and actions of the
        current period, and the model parameters ("params"). If target is "simulate",
        the function also depends on the dictionary of random keys ("keys"), which
        corresponds to the names of stochastic next functions.

    """
    targets = model.function_info.query("is_next").index.tolist()

    if target == Target.SOLVE:
        functions_dict = model.functions
    elif target == Target.SIMULATE:
        # For the simulation target, we need to extend the functions dictionary with
        # stochastic next states functions and their weights.
        functions_dict = _extend_functions_dict_for_simulation(model)
    else:
        raise ValueError(f"Invalid target: {target}")

    return concatenate_functions(
        functions=functions_dict,
        targets=targets,
        return_type="dict",
        enforce_signature=False,
    )


def get_next_stochastic_weights_function(
    model: InternalModel,
) -> Callable[..., dict[str, Array]]:
    """Get function that computes the weights for the next stochastic states.

    Args:
        model: Internal model instance.

    Returns:
        Function that computes the weights for the next stochastic states.

    """
    targets = [
        f"weight_{name}"
        for name in model.function_info.query("is_stochastic_next").index.tolist()
    ]

    return concatenate_functions(
        functions=model.functions,
        targets=targets,
        return_type="dict",
        enforce_signature=False,
    )


def _extend_functions_dict_for_simulation(
    model: InternalModel,
) -> dict[str, Callable[..., Scalar]]:
    """Extend the functions dictionary for the simulation target.

    Args:
        model: Internal model instance.

    Returns:
        Extended functions dictionary.

    """
    stochastic_targets = model.function_info.query("is_stochastic_next").index

    # Handle stochastic next states functions
    # ----------------------------------------------------------------------------------
    # We generate stochastic next states functions that simulate the next state given
    # a random key (think of a seed) and the weights corresponding to the labels of the
    # stochastic variable. The weights are computed using the stochastic weight
    # functions, which we add the to functions dict. `dags.concatenate_functions` then
    # generates a function that computes the weights and simulates the next state in
    # one go.
    # ----------------------------------------------------------------------------------
    stochastic_next = {
        name: _create_stochastic_next_func(
            name, labels=model.grids[name.removeprefix("next_")]
        )
        for name in stochastic_targets
    }

    stochastic_weights = {
        f"weight_{name}": model.functions[f"weight_{name}"]
        for name in stochastic_targets
    }

    # Overwrite model.functions with generated stochastic next states functions
    # ----------------------------------------------------------------------------------
    return model.functions | stochastic_next | stochastic_weights


def _create_stochastic_next_func(name: str, labels: Array) -> StochasticNextFunction:
    """Get function that simulates the next state of a stochastic variable.

    Args:
        name: Name of the stochastic variable.
        labels: 1d array of labels.

    Returns:
        A function that simulates the next state of the stochastic variable. The
        function must be called with keyword arguments:
        - weight_{name}: 2d array of weights. The first dimension corresponds to the
          number of simulation units. The second dimension corresponds to the number of
          grid points (labels).
        - keys: Dictionary with random key arrays. Dictionary keys correspond to the
          names of stochastic next functions, e.g. 'next_health'.

    """

    @with_signature(args=[f"weight_{name}", "keys"])
    def next_stochastic_state(keys: dict[str, Array], **kwargs: Array) -> Array:
        return random_choice(
            labels=labels,
            probs=kwargs[f"weight_{name}"],
            key=keys[name],
        )

    return next_stochastic_state



from collections import Counter
from collections.abc import Iterable
from itertools import chain
from typing import TypeVar

T = TypeVar("T")


def find_duplicates(*containers: Iterable[T]) -> set[T]:
    combined = chain.from_iterable(containers)
    counts = Counter(combined)
    return {v for v, count in counts.items() if count > 1}


def first_non_none(*args: T | None) -> T:
    """Return the first non-None argument.

    Args:
        *args: Arguments to check.

    Returns:
        The first non-None argument.

    Raises:
        ValueError: If all arguments are None.

    """
    for arg in args:
        if arg is not None:
            return arg
    raise ValueError("All arguments are None")



# Copyright 2019 The JAX Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Modifications made by Tim Mensinger, 2024

import functools
import itertools
import operator
from collections.abc import Sequence

import jax.numpy as jnp
from jax import Array, jit, lax, util


@jit
def map_coordinates(
    input: Array,
    coordinates: Sequence[Array],
) -> Array:
    """Map the input array to new coordinates using linear interpolation.

    Modified from JAX implementation of :func:`scipy.ndimage.map_coordinates`.

    Given an input array and a set of coordinates, this function returns the
    interpolated values of the input array at those coordinates. For coordinates outside
    the input array, linear extrapolation is used.

    Args:
      input: N-dimensional input array from which values are interpolated.
      coordinates: length-N sequence of arrays specifying the coordinates
        at which to evaluate the interpolated values

    Returns:
      The interpolated (extrapolated) values at the specified coordinates.

    """
    if len(coordinates) != input.ndim:
        raise ValueError(
            "coordinates must be a sequence of length input.ndim, but "
            f"{len(coordinates)} != {input.ndim}"
        )

    interpolation_data = [
        _compute_indices_and_weights(coordinate, size)
        for coordinate, size in util.safe_zip(coordinates, input.shape)
    ]

    interpolation_values = []
    for indices_and_weights in itertools.product(*interpolation_data):
        indices, weights = util.unzip2(indices_and_weights)
        contribution = input[indices]
        weighted_value = _multiply_all(weights) * contribution
        interpolation_values.append(weighted_value)

    result = _sum_all(interpolation_values)

    if jnp.issubdtype(input.dtype, jnp.integer):
        result = _round_half_away_from_zero(result)

    return result.astype(input.dtype)


def _compute_indices_and_weights(
    coordinate: Array, input_size: int
) -> list[tuple[Array, Array]]:
    """Compute indices and weights for linear interpolation."""
    lower_index = jnp.clip(jnp.floor(coordinate), 0, input_size - 2).astype(jnp.int32)
    upper_weight = coordinate - lower_index
    lower_weight = 1 - upper_weight
    return [(lower_index, lower_weight), (lower_index + 1, upper_weight)]


def _multiply_all(arrs: Sequence[Array]) -> Array:
    """Multiply all arrays in the sequence."""
    return functools.reduce(operator.mul, arrs)


def _sum_all(arrs: Sequence[Array]) -> Array:
    """Sum all arrays in the sequence."""
    return functools.reduce(operator.add, arrs)


def _round_half_away_from_zero(a: Array) -> Array:
    return a if jnp.issubdtype(a.dtype, jnp.integer) else lax.round(a)



import logging

import jax.numpy as jnp
from jax import Array

from lcm.interfaces import StateActionSpace
from lcm.typing import MaxQcOverDFunction, MaxQOverCFunction, ParamsDict


def solve(
    params: ParamsDict,
    state_action_spaces: dict[int, StateActionSpace],
    max_Q_over_c_functions: dict[int, MaxQOverCFunction],
    max_Qc_over_d_functions: dict[int, MaxQcOverDFunction],
    logger: logging.Logger,
) -> dict[int, Array]:
    """Solve a model using grid search.

    Args:
        params: Dict of model parameters.
        state_action_spaces: Dict with one state_action_space per period.
        max_Q_over_c_functions: Dict with one function per period. The functions
            calculate the maximum of the Q-function over the continuous actions. The
            result corresponds to the Qc-function of that period.
        max_Qc_over_d_functions: Dict with one function per period. The functions
            calculate the the (expected) maximum of the Qc-function over the discrete
            actions. The result corresponds to the value function array of that period.
        logger: Logger that logs to stdout.

    Returns:
        Dict with one value function array per period.

    """
    n_periods = len(state_action_spaces)
    solution = {}
    next_V_arr = jnp.empty(0)

    logger.info("Starting solution")

    # backwards induction loop
    for period in reversed(range(n_periods)):
        state_action_space = state_action_spaces[period]

        max_Qc_over_d = max_Qc_over_d_functions[period]
        max_Q_over_c = max_Q_over_c_functions[period]

        # evaluate Q-function on states and actions, and maximize over continuous
        # actions
        Qc_arr = max_Q_over_c(
            **state_action_space.states,
            **state_action_space.discrete_actions,
            **state_action_space.continuous_actions,
            next_V_arr=next_V_arr,
            params=params,
        )

        # maximize Qc-function evaluations over discrete actions
        V_arr = max_Qc_over_d(Qc_arr, params=params)

        solution[period] = V_arr
        next_V_arr = V_arr
        logger.info("Period: %s", period)

    return solution



"""Collection of classes that are used by the user to define the model and grids."""

import dataclasses as dc
from dataclasses import KW_ONLY, dataclass, field
from typing import Any

from lcm.exceptions import ModelInitilizationError, format_messages
from lcm.grids import Grid
from lcm.typing import UserFunction


@dataclass(frozen=True)
class Model:
    """A user model which can be processed into an internal model.

    Attributes:
        description: Description of the model.
        n_periods: Number of periods in the model.
        functions: Dictionary of user provided functions that define the functional
            relationships between model variables. It must include at least a function
            called 'utility'.
        actions: Dictionary of user provided actions.
        states: Dictionary of user provided states.

    """

    description: str | None = None
    _: KW_ONLY
    n_periods: int
    functions: dict[str, UserFunction] = field(default_factory=dict)
    actions: dict[str, Grid] = field(default_factory=dict)
    states: dict[str, Grid] = field(default_factory=dict)

    def __post_init__(self) -> None:
        _validate_attribute_types(self)
        _validate_logical_consistency(self)

    def replace(self, **kwargs: Any) -> "Model":  # noqa: ANN401
        """Replace the attributes of the model.

        Args:
            **kwargs: Keyword arguments to replace the attributes of the model.

        Returns:
            A new model with the replaced attributes.

        """
        try:
            return dc.replace(self, **kwargs)
        except TypeError as e:
            raise ModelInitilizationError(
                f"Failed to replace attributes of the model. The error was: {e}"
            ) from e


def _validate_attribute_types(model: Model) -> None:  # noqa: C901
    """Validate the types of the model attributes."""
    error_messages = []

    # Validate types of states and actions
    # ----------------------------------------------------------------------------------
    for attr_name in ("actions", "states"):
        attr = getattr(model, attr_name)
        if isinstance(attr, dict):
            for k, v in attr.items():
                if not isinstance(k, str):
                    error_messages.append(f"{attr_name} key {k} must be a string.")
                if not isinstance(v, Grid):
                    error_messages.append(f"{attr_name} value {v} must be an LCM grid.")
        else:
            error_messages.append(f"{attr_name} must be a dictionary.")

    # Validate types of functions
    # ----------------------------------------------------------------------------------
    if isinstance(model.functions, dict):
        for k, v in model.functions.items():
            if not isinstance(k, str):
                error_messages.append(f"function keys must be a strings, but is {k}.")
            if not callable(v):
                error_messages.append(
                    f"function values must be a callable, but is {v}."
                )
    else:
        error_messages.append("functions must be a dictionary.")

    if error_messages:
        msg = format_messages(error_messages)
        raise ModelInitilizationError(msg)


def _validate_logical_consistency(model: Model) -> None:
    """Validate the logical consistency of the model."""
    error_messages = []

    if model.n_periods < 1:
        error_messages.append("Number of periods must be a positive integer.")

    if "utility" not in model.functions:
        error_messages.append(
            "Utility function is not defined. LCM expects a function called 'utility' "
            "in the functions dictionary.",
        )

    states_without_next_func = [
        state for state in model.states if f"next_{state}" not in model.functions
    ]
    if states_without_next_func:
        error_messages.append(
            "Each state must have a corresponding next state function. For the "
            "following states, no next state function was found: "
            f"{states_without_next_func}.",
        )

    states_and_actions_overlap = set(model.states) & set(model.actions)
    if states_and_actions_overlap:
        error_messages.append(
            "States and actions cannot have overlapping names. The following names "
            f"are used in both states and actions: {states_and_actions_overlap}.",
        )

    if error_messages:
        msg = format_messages(error_messages)
        raise ModelInitilizationError(msg)



import logging


def get_logger(*, debug_mode: bool) -> logging.Logger:
    """Get a logger that logs to stdout.

    Args:
        debug_mode: Whether to log debug messages.

    Returns:
        Logger that logs to stdout.

    """
    logging.basicConfig(level=logging.WARNING)
    logger = logging.getLogger("lcm")

    if debug_mode:
        logger.setLevel(logging.DEBUG)

    return logger



import functools
import inspect
from copy import deepcopy
from typing import cast

import pandas as pd
from dags.signature import with_signature
from jax import Array

from lcm.functools import all_as_args
from lcm.input_processing.create_params_template import create_params_template
from lcm.input_processing.util import (
    get_function_info,
    get_grids,
    get_gridspecs,
    get_variable_info,
)
from lcm.interfaces import InternalModel
from lcm.typing import InternalUserFunction, ParamsDict, Scalar, ShockType, UserFunction
from lcm.user_model import Model


def process_model(model: Model) -> InternalModel:
    """Process the user model.

    This entails the following steps:

    - Set defaults where needed
    - Generate derived information
    - Check that the model specification is valid.

    Args:
        model: The model as provided by the user.

    Returns:
        The processed model.

    """
    params = create_params_template(model)

    return InternalModel(
        grids=get_grids(model),
        gridspecs=get_gridspecs(model),
        variable_info=get_variable_info(model),
        functions=_get_internal_functions(model, params=params),
        function_info=get_function_info(model),
        params=params,
        # currently no additive utility shocks are supported
        random_utility_shocks=ShockType.NONE,
        n_periods=model.n_periods,
    )


def _get_internal_functions(
    model: Model,
    params: ParamsDict,
) -> dict[str, InternalUserFunction]:
    """Process the user provided model functions.

    Args:
        model: The model as provided by the user.
        params: The parameters of the model.

    Returns:
        Dictionary containing all functions of the model. The keys are the names of the
        functions. The values are the processed functions. The main difference between
        processed and unprocessed functions is that processed functions take `params` as
        argument.

    """
    variable_info = get_variable_info(model)
    grids = get_grids(model)

    raw_functions = deepcopy(model.functions)

    for var in model.states:
        if variable_info.loc[var, "is_stochastic"]:
            raw_functions[f"next_{var}"] = _get_stochastic_next_function(
                raw_func=raw_functions[f"next_{var}"],
                grid=grids[var],
            )

            raw_functions[f"weight_next_{var}"] = _get_stochastic_weight_function(
                raw_func=raw_functions[f"next_{var}"],
                name=var,
                variable_info=variable_info,
            )

    # ==================================================================================
    # Add 'params' argument to functions
    # ==================================================================================
    # We wrap the user functions such that they can be called with the 'params' argument
    # instead of the individual parameters. This is done for all functions except for
    # the dynamically generated weighting functions for stochastic next functions, since
    # they are constructed to accept the 'params' argument by default.

    functions: dict[str, InternalUserFunction] = {}

    for func_name, func in raw_functions.items():
        is_weight_next_function = func_name.startswith("weight_next_")

        if is_weight_next_function:
            processed_func = cast(InternalUserFunction, func)

        # params[name] contains the dictionary of parameters for the function, which
        # is empty if the function does not depend on any model parameters.
        elif params[func_name]:
            processed_func = _replace_func_parameters_by_params(
                func=func,
                params=params,
                name=func_name,
            )

        else:
            processed_func = _add_dummy_params_argument(func)

        functions[func_name] = processed_func

    return functions


def _replace_func_parameters_by_params(
    func: UserFunction, params: ParamsDict, name: str
) -> InternalUserFunction:
    old_signature = list(inspect.signature(func).parameters)
    new_kwargs = [p for p in old_signature if p not in params[name]] + ["params"]

    @with_signature(args=new_kwargs)
    @functools.wraps(func)
    def processed_func(*args: Scalar, params: ParamsDict, **kwargs: Scalar) -> Scalar:
        return func(*args, **kwargs, **params[name])

    return processed_func


def _add_dummy_params_argument(func: UserFunction) -> InternalUserFunction:
    old_signature = list(inspect.signature(func).parameters)

    new_kwargs = [*old_signature, "params"]

    @with_signature(args=new_kwargs)
    @functools.wraps(func)
    def processed_func(*args: Scalar, params: ParamsDict, **kwargs: Scalar) -> Scalar:  # noqa: ARG001
        return func(*args, **kwargs)

    return processed_func


def _get_stochastic_next_function(raw_func: UserFunction, grid: Array) -> UserFunction:
    @functools.wraps(raw_func)
    def next_func(*args: Scalar, **kwargs: Scalar) -> Array:  # noqa: ARG001
        return grid

    return next_func


def _get_stochastic_weight_function(
    raw_func: UserFunction, name: str, variable_info: pd.DataFrame
) -> InternalUserFunction:
    """Get a function that returns the transition weights of a stochastic variable.

    Example:
    Consider a stochastic variable 'health' that takes two values {0, 1}. The transition
    matrix is thus 2x2. We create the weighting function and then select the weights
    that correspond to the case where 'health' is 0.

    >>> from lcm.mark import StochasticInfo
    >>> def next_health(health):
    >>>     pass
    >>> next_health._stochastic_info = StochasticInfo()
    >>> params = {"shocks": {"health": np.arange(4).reshape(2, 2)}}
    >>> weight_func = _get_stochastic_weight_function(
    >>>     raw_func=next_health,
    >>>     name="health"
    >>>     variable_info=variable_info,
    >>>     grids=grids,
    >>> )
    >>> weight_func(health=0, params=params)
    >>> array([0, 1])


    Args:
        raw_func: The raw next function of the stochastic variable.
        name: The name of the stochastic variable.
        variable_info: A table with information about model variables.

    Returns:
        A function that returns the transition weights of the stochastic variable.

    """
    function_parameters = list(inspect.signature(raw_func).parameters)

    # Assert that stochastic next function only depends on discrete variables or period
    invalid = {
        arg
        for arg in function_parameters
        if arg != "_period" and not variable_info.loc[arg, "is_discrete"]
    }

    if invalid:
        raise ValueError(
            "Stochastic variables can only depend on discrete variables and '_period', "
            f"but {name} depends on {invalid}.",
        )

    new_kwargs = [*function_parameters, "params"]

    @with_signature(args=new_kwargs)
    def weight_func(*args: Array, params: ParamsDict, **kwargs: Array) -> Array:
        args = all_as_args(args, kwargs, arg_names=new_kwargs)
        return params["shocks"][name][*args]

    return cast(InternalUserFunction, weight_func)



from collections.abc import Callable
from typing import Any

import jax.numpy as jnp
from dags import concatenate_functions
from dags.signature import with_signature
from jax import Array

from lcm.dispatchers import productmap
from lcm.function_representation import get_value_function_representation
from lcm.functools import get_union_of_arguments
from lcm.interfaces import InternalModel, StateSpaceInfo
from lcm.next_state import get_next_state_function, get_next_stochastic_weights_function
from lcm.typing import InternalUserFunction, ParamsDict, Scalar, Target


def get_Q_and_F(
    model: InternalModel,
    next_state_space_info: StateSpaceInfo,
    period: int,
) -> Callable[..., tuple[Array, Array]]:
    """Get the state-action (Q) and feasibility (F) function for a given period.

    Args:
        model: The internal model object.
        next_state_space_info: The state space information of the next period.
        period: The current period.

    Returns:
        A function that computes the state-action values (Q) and the feasibilities (F)
        for the given period.

    """
    is_last_period = period == model.n_periods - 1

    if is_last_period:
        Q_and_F = get_Q_and_F_terminal(model, period=period)
    else:
        Q_and_F = get_Q_and_F_non_terminal(
            model, next_state_space_info=next_state_space_info, period=period
        )

    return Q_and_F


def get_Q_and_F_non_terminal(
    model: InternalModel,
    next_state_space_info: StateSpaceInfo,
    period: int,
) -> Callable[..., tuple[Array, Array]]:
    """Get the state-action (Q) and feasibility (F) function for a non-terminal period.

    Args:
        model: The internal model object.
        next_state_space_info: The state space information of the next period.
        period: The current period.

    Returns:
        A function that computes the state-action values (Q) and the feasibilities (F)
        for a non-terminal period.

    """
    stochastic_variables = model.variable_info.query("is_stochastic").index.tolist()

    # ----------------------------------------------------------------------------------
    # Generate dynamic functions
    # ----------------------------------------------------------------------------------

    # Function required to calculate instantaneous utility and feasibility
    U_and_F = _get_U_and_F(model)

    # Functions required to calculate the expected continuation values
    state_transition = get_next_state_function(model, target=Target.SOLVE)
    next_stochastic_states_weights = get_next_stochastic_weights_function(model)
    joint_weights_from_marginals = _get_joint_weights_function(stochastic_variables)
    _scalar_next_V = get_value_function_representation(next_state_space_info)
    next_V = productmap(
        _scalar_next_V,
        variables=tuple(f"next_{var}" for var in stochastic_variables),
    )

    # ----------------------------------------------------------------------------------
    # Create the state-action value and feasibility function
    # ----------------------------------------------------------------------------------
    arg_names_of_Q_and_F = _get_arg_names_of_Q_and_F(
        [U_and_F, state_transition, next_stochastic_states_weights],
        include={"params", "next_V_arr"},
        exclude={"_period"},
    )

    @with_signature(args=arg_names_of_Q_and_F)
    def Q_and_F(
        params: ParamsDict, next_V_arr: Array, **states_and_actions: Scalar
    ) -> tuple[Scalar, Scalar]:
        """Calculate the state-action value and feasibility for a non-terminal period.

        Args:
            params: The parameters.
            next_V_arr: The next period's value function array.
            **states_and_actions: The current states and actions.

        Returns:
            A tuple containing the arrays with state-action values and feasibilities.

        """
        # ------------------------------------------------------------------------------
        # Calculate the expected continuation values
        # ------------------------------------------------------------------------------
        next_states = state_transition(
            **states_and_actions,
            _period=period,
            params=params,
        )

        marginal_next_stochastic_states_weights = next_stochastic_states_weights(
            **states_and_actions,
            _period=period,
            params=params,
        )

        joint_next_stochastic_states_weights = joint_weights_from_marginals(
            **marginal_next_stochastic_states_weights
        )

        # As we productmap'd the value function over the stochastic variables, the
        # resulting next value function gets a new dimension for each stochastic
        # variable.
        next_V_at_stochastic_states_arr = next_V(**next_states, next_V_arr=next_V_arr)

        # We then take the weighted average of the next value function at the stochastic
        # states to get the expected next value function.
        next_V_expected_arr = jnp.average(
            next_V_at_stochastic_states_arr,
            weights=joint_next_stochastic_states_weights,
        )

        # ------------------------------------------------------------------------------
        # Calculate the instantaneous utility and feasibility
        # ------------------------------------------------------------------------------
        U_arr, F_arr = U_and_F(
            **states_and_actions,
            _period=period,
            params=params,
        )

        Q_arr = U_arr + params["beta"] * next_V_expected_arr

        return Q_arr, F_arr

    return Q_and_F  # type: ignore[return-value]


def get_Q_and_F_terminal(
    model: InternalModel,
    period: int,
) -> Callable[..., tuple[Array, Array]]:
    """Get the state-action (Q) and feasibility (F) function for the terminal period.

    Args:
        model: The internal model object.
        period: The current period.

    Returns:
        A function that computes the state-action values (Q) and the feasibilities (F)
        for the terminal period.

    """
    U_and_F = _get_U_and_F(model)

    arg_names_of_Q_and_F = _get_arg_names_of_Q_and_F(
        [U_and_F],
        # While the terminal period does not depend on the value function array, we
        # include it in the signature, such that we can treat all periods uniformly
        # during the solution and simulation.
        include={"params", "next_V_arr"},
        exclude={"_period"},
    )

    @with_signature(args=arg_names_of_Q_and_F)
    def Q_and_F(
        params: ParamsDict,
        next_V_arr: Array,  # noqa: ARG001
        **states_and_actions: Scalar,
    ) -> tuple[Scalar, Scalar]:
        """Calculate the state-action values and feasibilities for the terminal period.

        Args:
            params: The parameters.
            next_V_arr: The next period's value function array (unused here).
            **states_and_actions: The current states and actions.

        Returns:
            A tuple containing the arrays with state-action values and feasibilities.

        """
        return U_and_F(
            **states_and_actions,
            _period=period,
            params=params,
        )

    return Q_and_F  # type: ignore[return-value]


# ======================================================================================
# Helper functions
# ======================================================================================


def _get_arg_names_of_Q_and_F(
    deps: list[Callable[..., Any]],
    include: set[str] = set(),  # noqa: B006
    exclude: set[str] = set(),  # noqa: B006
) -> list[str]:
    """Get the argument names of the dependencies.

    Args:
        deps: List of dependencies.
        include: Set of argument names to include.
        exclude: Set of argument names to exclude.

    Returns:
        The union of the argument names in deps and include, except for those in
        exclude.

    """
    deps_arg_names = get_union_of_arguments(deps)
    return list(include | deps_arg_names - exclude)


def _get_joint_weights_function(
    stochastic_variables: list[str],
) -> Callable[..., Array]:
    """Get function that calculates the joint weights.

    This function takes the weights of the individual stochastic variables and
    multiplies them together to get the joint weights on the product space of the
    stochastic variables.

    Args:
        stochastic_variables: List of stochastic variables.

    Returns:
        A function that computes the outer product of the weights of the stochastic
        variables.

    """
    arg_names = [f"weight_next_{var}" for var in stochastic_variables]

    @with_signature(args=arg_names)
    def _outer(**kwargs: Array) -> Array:
        weights = jnp.array(list(kwargs.values()))
        return jnp.prod(weights)

    return productmap(_outer, variables=tuple(arg_names))


def _get_U_and_F(model: InternalModel) -> Callable[..., tuple[Scalar, Scalar]]:
    """Get the instantaneous utility and feasibility function.

    Note:
    -----
    U may depend on all kinds of other functions (taxes, transfers, ...), which will be
    executed if they matter for the value of U.

    Args:
        model: The internal model object.

    Returns:
        The instantaneous utility and feasibility function.

    """
    functions = {"feasibility": _get_feasibility(model), **model.functions}
    return concatenate_functions(
        functions=functions,
        targets=["utility", "feasibility"],
        enforce_signature=False,
    )


def _get_feasibility(model: InternalModel) -> InternalUserFunction:
    """Create a function that combines all constraint functions into a single one.

    Args:
        model: The internal model object.

    Returns:
        The combined constraint function (feasibility).

    """
    constraints = model.function_info.query("is_constraint").index.tolist()

    if constraints:
        combined_constraint = concatenate_functions(
            functions=model.functions,
            targets=constraints,
            aggregator=jnp.logical_and,
        )
    else:

        def combined_constraint(**kwargs: Scalar) -> bool:  # noqa: ARG001
            """Dummy feasibility function that always returns True."""
            return True

    return combined_constraint



import pandas as pd
from dags import get_ancestors
from jax import Array

from lcm.grids import ContinuousGrid, Grid
from lcm.typing import UserFunction
from lcm.user_model import Model


def get_function_info(model: Model) -> pd.DataFrame:
    """Derive information about functions in the model.

    Args:
        model: The model as provided by the user.

    Returns:
        A table with information about all functions in the model. The index contains
        the name of a model function. The columns are booleans that are True if the
        function has the corresponding property. The columns are: is_next,
        is_stochastic_next, is_constraint.

    """
    info = pd.DataFrame(index=list(model.functions))
    # Convert both filter and constraint to constraints, until we forbid filters.
    info["is_constraint"] = info.index.str.endswith(("_constraint", "_filter"))
    info["is_next"] = info.index.str.startswith("next_") & ~info["is_constraint"]
    info["is_stochastic_next"] = [
        hasattr(func, "_stochastic_info") and info.loc[func_name]["is_next"]
        for func_name, func in model.functions.items()
    ]
    return info


def get_variable_info(model: Model) -> pd.DataFrame:
    """Derive information about all variables in the model.

    Args:
        model: The model as provided by the user.

    Returns:
        A table with information about all variables in the model. The index contains
        the name of a model variable. The columns are booleans that are True if the
        variable has the corresponding property. The columns are: is_state, is_action,
        is_continuous, is_discrete.

    """
    function_info = get_function_info(model)

    variables = model.states | model.actions

    info = pd.DataFrame(index=list(variables))

    info["is_state"] = info.index.isin(model.states)
    info["is_action"] = ~info["is_state"]

    info["is_continuous"] = [
        isinstance(spec, ContinuousGrid) for spec in variables.values()
    ]
    info["is_discrete"] = ~info["is_continuous"]

    info["is_stochastic"] = [
        (var in model.states and function_info.loc[f"next_{var}", "is_stochastic_next"])
        for var in variables
    ]

    auxiliary_variables = _get_auxiliary_variables(
        state_variables=info.query("is_state").index.tolist(),
        function_info=function_info,
        user_functions=model.functions,
    )
    info["is_auxiliary"] = [var in auxiliary_variables for var in variables]

    order = info.query("is_discrete & is_state").index.tolist()
    order += info.query("is_discrete & is_action").index.tolist()
    order += info.query("is_continuous & is_state").index.tolist()
    order += info.query("is_continuous & is_action").index.tolist()

    if set(order) != set(info.index):
        raise ValueError("Order and index do not match.")

    return info.loc[order]


def _get_auxiliary_variables(
    state_variables: list[str],
    function_info: pd.DataFrame,
    user_functions: dict[str, UserFunction],
) -> list[str]:
    """Get state variables that only occur in next functions.

    Args:
        state_variables: List of state variable names.
        function_info: A table with information about all
            functions in the model. The index contains the name of a function. The
            columns are booleans that are True if the function has the corresponding
            property. The columns are: is_filter, is_constraint, is_next.
        user_functions: Dictionary that maps names of functions to functions.

    Returns:
        List of state variable names that are only used in next functions.

    """
    non_next_functions = function_info.query("~is_next").index.tolist()
    user_functions = {name: user_functions[name] for name in non_next_functions}
    ancestors = get_ancestors(
        user_functions,
        targets=list(user_functions),
        include_targets=True,
    )
    return list(set(state_variables).difference(set(ancestors)))


def get_gridspecs(
    model: Model,
) -> dict[str, Grid]:
    """Create a dictionary of grid specifications for each variable in the model.

    Args:
        model (dict): The model as provided by the user.

    Returns:
        Dictionary containing all variables of the model. The keys are the names of the
        variables. The values describe which values the variable can take. For discrete
        variables these are the codes. For continuous variables this is information
        about how to build the grids.

    """
    variable_info = get_variable_info(model)

    raw_variables = model.states | model.actions
    order = variable_info.index.tolist()
    return {k: raw_variables[k] for k in order}


def get_grids(
    model: Model,
) -> dict[str, Array]:
    """Create a dictionary of array grids for each variable in the model.

    Args:
        model: The model as provided by the user.

    Returns:
        Dictionary containing all variables of the model. The keys are the names of the
        variables. The values are the grids.

    """
    variable_info = get_variable_info(model)
    gridspecs = get_gridspecs(model)

    grids = {name: spec.to_jax() for name, spec in gridspecs.items()}
    order = variable_info.index.tolist()
    return {k: grids[k] for k in order}



import logging
from collections.abc import Callable
from functools import partial

import jax
import jax.numpy as jnp
import pandas as pd
from jax import Array, vmap

from lcm.dispatchers import simulation_spacemap, vmap_1d
from lcm.interfaces import (
    InternalModel,
    InternalSimulationPeriodResults,
    StateActionSpace,
)
from lcm.max_Qc_over_d import get_argmax_and_max_Qc_over_d
from lcm.random import draw_random_seed, generate_simulation_keys
from lcm.simulation.processing import as_panel, process_simulated_data
from lcm.state_action_space import create_state_action_space
from lcm.typing import ArgmaxQOverCFunction, ParamsDict


def solve_and_simulate(
    params: ParamsDict,
    initial_states: dict[str, Array],
    argmax_and_max_Q_over_c_functions: dict[int, ArgmaxQOverCFunction],
    model: InternalModel,
    next_state: Callable[..., dict[str, Array]],
    logger: logging.Logger,
    solve_model: Callable[..., dict[int, Array]],
    *,
    additional_targets: list[str] | None = None,
    seed: int | None = None,
) -> pd.DataFrame:
    """First solve the model and then simulate the model forward in time.

    Same docstring as `simulate` mutatis mutandis.

    """
    V_arr_dict = solve_model(params)
    return simulate(
        params=params,
        initial_states=initial_states,
        argmax_and_max_Q_over_c_functions=argmax_and_max_Q_over_c_functions,
        model=model,
        next_state=next_state,
        logger=logger,
        V_arr_dict=V_arr_dict,
        additional_targets=additional_targets,
        seed=seed,
    )


def simulate(
    params: ParamsDict,
    initial_states: dict[str, Array],
    argmax_and_max_Q_over_c_functions: dict[int, ArgmaxQOverCFunction],
    model: InternalModel,
    next_state: Callable[..., dict[str, Array]],
    logger: logging.Logger,
    V_arr_dict: dict[int, Array],
    *,
    additional_targets: list[str] | None = None,
    seed: int | None = None,
) -> pd.DataFrame:
    """Simulate the model forward in time given pre-computed value function arrays.

    Args:
        params: Dict of model parameters.
        initial_states: List of initial states to start from. Typically from the
            observed dataset.
        argmax_and_max_Q_over_c_functions: Dict of functions of length n_periods. Each
            function calculates the argument maximizing Q over the continuous actions.
        next_state: Function that returns the next state given the current
            state and action variables. For stochastic variables, it returns a random
            draw from the distribution of the next state.
        model: Model instance.
        logger: Logger that logs to stdout.
        V_arr_dict: Dict of value function arrays of length n_periods.
        additional_targets: List of targets to compute. If provided, the targets
            are computed and added to the simulation results.
        seed: Random number seed; will be passed to `jax.random.key`. If not provided,
            a random seed will be generated.

    Returns:
        DataFrame with the simulation results.

    """
    if seed is None:
        seed = draw_random_seed()

    logger.info("Starting simulation")

    # Preparations
    # ----------------------------------------------------------------------------------
    n_periods = len(V_arr_dict)
    n_initial_states = len(next(iter(initial_states.values())))

    state_action_space = create_state_action_space(
        model=model,
        initial_states=initial_states,
    )

    argmax_and_max_Qc_over_d = get_argmax_and_max_Qc_over_d(
        variable_info=model.variable_info
    )

    # The following variables are updated during the forward simulation
    states = initial_states
    key = jax.random.key(seed=seed)

    # Forward simulation
    # ----------------------------------------------------------------------------------
    simulation_results = {}

    for period in range(n_periods):
        state_action_space = state_action_space.replace(states)

        discrete_actions_grid_shape = tuple(
            len(grid) for grid in state_action_space.discrete_actions.values()
        )
        continuous_actions_grid_shape = tuple(
            len(grid) for grid in state_action_space.continuous_actions.values()
        )

        # Compute optimal continuous actions conditional on discrete actions
        # ------------------------------------------------------------------------------
        # We need to pass the value function array of the next period to the
        # argmax_and_max_Q_over_c function, as the current Q-function requires the next
        # periods's value funciton. In the last period, we pass an empty array.
        next_V_arr = V_arr_dict.get(period + 1, jnp.empty(0))

        argmax_and_max_Q_over_c = simulation_spacemap(
            argmax_and_max_Q_over_c_functions[period],
            actions_names=tuple(state_action_space.discrete_actions),
            states_names=tuple(state_action_space.states),
        )

        # Returns the optimal continuous action index conditional on the states and
        # discrete actions, as well as the maximum value.
        indices_argmax_Q_over_c, Qc_arr = argmax_and_max_Q_over_c(
            **state_action_space.states,
            **state_action_space.discrete_actions,
            **state_action_space.continuous_actions,
            next_V_arr=next_V_arr,
            params=params,
        )

        # The Qc-function values contain the information of how much value each discrete
        # action combination is worth, assuming the corresponding optimal continuous
        # actions are taken. To find the optimal discrete action, we therefore only need
        # to maximize the Qc-function values over the discrete actions.
        # ------------------------------------------------------------------------------
        indices_optimal_discrete_actions, V_arr = argmax_and_max_Qc_over_d(
            Qc_arr, params=params
        )

        # Look up the continuous actions index from the above set given the optimal
        # discrete actions.
        # ------------------------------------------------------------------------------
        indices_optimal_continuous_actions = _lookup_optimal_continuous_actions(
            indices_argmax_Q_over_c=indices_argmax_Q_over_c,
            discrete_argmax=indices_optimal_discrete_actions,
            discrete_actions_grid_shape=discrete_actions_grid_shape,
        )

        # Convert action indices to action values
        # ------------------------------------------------------------------------------
        optimal_actions = _lookup_actions_from_indices(
            indices_optimal_discrete_actions=indices_optimal_discrete_actions,
            indices_optimal_continuous_actions=indices_optimal_continuous_actions,
            discrete_actions_grid_shape=discrete_actions_grid_shape,
            continuous_actions_grid_shape=continuous_actions_grid_shape,
            state_action_space=state_action_space,
        )

        # Store results
        # ------------------------------------------------------------------------------
        simulation_results[period] = InternalSimulationPeriodResults(
            value=V_arr,
            actions=optimal_actions,
            states=states,
        )

        # Update states
        # ------------------------------------------------------------------------------
        key, stochastic_variables_keys = generate_simulation_keys(
            key=key,
            ids=model.function_info.query("is_stochastic_next").index.tolist(),
        )

        states_with_next_prefix = next_state(
            **states,
            **optimal_actions,
            _period=jnp.repeat(period, n_initial_states),
            params=params,
            keys=stochastic_variables_keys,
        )
        # 'next_' prefix is added by the next_state function, but needs to be removed
        # because in the next period, next states will be current states.
        states = {
            k.removeprefix("next_"): v for k, v in states_with_next_prefix.items()
        }

        logger.info("Period: %s", period)

    processed = process_simulated_data(
        simulation_results,
        model=model,
        params=params,
        additional_targets=additional_targets,
    )

    return as_panel(processed, n_periods=n_periods)


@partial(vmap_1d, variables=("indices_argmax_Q_over_c", "discrete_argmax"))
def _lookup_optimal_continuous_actions(
    indices_argmax_Q_over_c: Array,
    discrete_argmax: Array,
    discrete_actions_grid_shape: tuple[int, ...],
) -> Array:
    """Look up the optimal continuous action index given index of discrete action.

    Args:
        indices_argmax_Q_over_c: Index array of optimal continous actions conditional on
            discrete actions and states.
        discrete_argmax: Index array of optimal discrete actions.
        discrete_actions_grid_shape: Shape of the discrete actions grid.

    Returns:
        Index array of optimal continuous actions.

    """
    indices = jnp.unravel_index(discrete_argmax, shape=discrete_actions_grid_shape)
    return indices_argmax_Q_over_c[indices]


def _lookup_actions_from_indices(
    indices_optimal_discrete_actions: Array,
    indices_optimal_continuous_actions: Array,
    discrete_actions_grid_shape: tuple[int, ...],
    continuous_actions_grid_shape: tuple[int, ...],
    state_action_space: StateActionSpace,
) -> dict[str, Array]:
    """Lookup optimal actions from indices.

    Args:
        indices_optimal_discrete_actions: Indices of optimal discrete actions.
        indices_optimal_continuous_actions: Indices of optimal continuous actions.
        discrete_actions_grid_shape: Shape of the discrete actions grid.
        continuous_actions_grid_shape: Shape of the continuous actions grid.
        state_action_space: StateActionSpace instance.

    Returns:
        Dictionary of optimal actions.

    """
    optimal_discrete_actions = _lookup_values_from_indices(
        flat_indices=indices_optimal_discrete_actions,
        grids=state_action_space.discrete_actions,
        grids_shapes=discrete_actions_grid_shape,
    )

    optimal_continuous_actions = _lookup_values_from_indices(
        flat_indices=indices_optimal_continuous_actions,
        grids=state_action_space.continuous_actions,
        grids_shapes=continuous_actions_grid_shape,
    )

    return optimal_discrete_actions | optimal_continuous_actions


def _lookup_values_from_indices(
    flat_indices: Array,
    grids: dict[str, Array],
    grids_shapes: tuple[int, ...],
) -> dict[str, Array]:
    """Retrieve values from indices.

    Args:
        flat_indices: General indices. Represents the index of the flattened grid.
        grids: Dictionary of grid values.
        grids_shapes: Shape of the grids. Is used to unravel the index.

    Returns:
        Dictionary of values.

    """
    nd_indices = vmapped_unravel_index(flat_indices, grids_shapes)
    return {
        name: grid[index]
        for (name, grid), index in zip(grids.items(), nd_indices, strict=True)
    }


# vmap jnp.unravel_index over the first axis of the `indices` argument, while holding
# the `shape` argument constant (in_axes = (0, None)).
vmapped_unravel_index = vmap(jnp.unravel_index, in_axes=(0, None))



import inspect

import jax.numpy as jnp
import pandas as pd
from dags import concatenate_functions
from jax import Array

from lcm.dispatchers import vmap_1d
from lcm.interfaces import InternalModel, InternalSimulationPeriodResults
from lcm.typing import InternalUserFunction, ParamsDict


def process_simulated_data(
    results: dict[int, InternalSimulationPeriodResults],
    model: InternalModel,
    params: ParamsDict,
    additional_targets: list[str] | None = None,
) -> dict[str, Array]:
    """Process and flatten the simulation results.

    This function produces a dict of arrays for each var with dimension (n_periods *
    n_initial_states,). The arrays are flattened, so that the resulting dictionary has a
    one-dimensional array for each variable. The length of this array is the number of
    periods times the number of initial states. The order of array elements is given by
    an outer level of periods and an inner level of initial states ids.

    Args:
        results: Dict with simulation results. Each dict contains the value,
            actions, and states for one period. Actions and states are stored in a
            nested dictionary.
        model: Model.
        params: Parameters.
        additional_targets: List of additional targets to compute.

    Returns:
        Dict with processed simulation results. The keys are the variable names and the
        values are the flattened arrays, with dimension (n_periods * n_initial_states,).
        Additionally, the _period variable is added.

    """
    n_periods = len(results)
    n_initial_states = len(results[0].value)

    list_of_dicts = [
        {"value": d.value, **d.actions, **d.states} for d in results.values()
    ]
    dict_of_lists = {
        key: [d[key] for d in list_of_dicts] for key in list(list_of_dicts[0])
    }
    out = {key: jnp.concatenate(values) for key, values in dict_of_lists.items()}
    out["_period"] = jnp.repeat(jnp.arange(n_periods), n_initial_states)

    if additional_targets is not None:
        calculated_targets = _compute_targets(
            out,
            targets=additional_targets,
            model_functions=model.functions,
            params=params,
        )
        out = {**out, **calculated_targets}

    return out


def as_panel(processed: dict[str, Array], n_periods: int) -> pd.DataFrame:
    """Convert processed simulation results to panel.

    Args:
        processed: Dict with processed simulation results.
        n_periods: Number of periods.

    Returns:
        Panel with the simulation results. The index is a multi-index with the first
        level corresponding to the initial state id and the second level corresponding
        to the period. The columns correspond to the value, and the action and state
        variables, and potentially auxiliary variables.

    """
    n_initial_states = len(processed["value"]) // n_periods
    index = pd.MultiIndex.from_product(
        [range(n_periods), range(n_initial_states)],
        names=["period", "initial_state_id"],
    )
    return pd.DataFrame(processed, index=index)


def _compute_targets(
    processed_results: dict[str, Array],
    targets: list[str],
    model_functions: dict[str, InternalUserFunction],
    params: ParamsDict,
) -> dict[str, Array]:
    """Compute targets.

    Args:
        processed_results: Dict with processed simulation results. Values must be
            one-dimensional arrays.
        targets: List of targets to compute.
        model_functions: Dict with model functions.
        params: Dict with model parameters.

    Returns:
        Dict with computed targets.

    """
    target_func = concatenate_functions(
        functions=model_functions,
        targets=targets,
        return_type="dict",
    )

    # get list of variables over which we want to vectorize the target function
    variables = tuple(
        p for p in list(inspect.signature(target_func).parameters) if p != "params"
    )

    target_func = vmap_1d(target_func, variables=variables)

    kwargs = {k: v for k, v in processed_results.items() if k in variables}
    return target_func(params=params, **kwargs)
